# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
satellites
----------

This is a sample dataset build into the engine, this simplifies a few things:

- We can write test scripts using this data, knowing that it will always be available.
- We can write examples using this data, knowing the results will always match.

This data was obtained from:
https://github.com/devstronomy/nasa-data-scraper/blob/master/data/json/satellites.json

Licence @ 02-JAN-2022 when copied - MIT Licences attested, but data appears to be
from NASA, which is Public Domain.

To access this dataset you can either run a query against dataset :satellites: or you
can instantiate a SatelliteData() class and use it like a Relation.

This has a companion dataset, $planets, to help test joins.
"""


from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import OrsoTypes


def read(*args):
    import base64
    import io

    import pyarrow.parquet as pq

    """The table is saved parquet table, base85 encoded."""

    return pq.read_table(
        io.BytesIO(
            base64.b85decode(
                b"P(e~L6$BNK78RZZOcmk+6$BCh04TLD{a}a$EeHUH);=8&kgpv83;+xO3;+xO3_!^xmt1nmC6`=s$tBY%GXOOJIRLwI-KsS!)~i;lQk^O_D%7V=n=)OhG%3=fMvD?1Dl{n2pFVr?+^I7s&YL!C(wr$XCKXH&!2=B}kiY>23=qJ-e*5y>t2Zy+yLRi+ohvsk+_!GqvR$h-E!wkY%aR={HZ0h$UIA~ku|^talrcsaUv#lW7FSd;MHEjou|yI_6fr~)KlHFe4mZ>=LkusputEwalrTaFA9S!m1{VbXFv0xN%PzUxQp+r{ywb`lshm>ED4~4P$tIaxQpqHdJkrP_i5ybMAc6eR#~yjyQO6u{1c0zq0ssR66#x~20~M45EEVDc6$BLq6$Ts_2(bVF0000002l}X000000000M04NBt000000000O2mt^90000000000D77#BAi)U+00RI30OA4x2rmEu0s{mE1_uZU3JVMk4i69!5)%{^78e*98XFuP9v>hfA|oUvCMPH<Dl054E-x@KGBY$aHa9pqIy*c)K0iP~LPJDFMn_0VN=r;lPESx#Qd3k_R##YAT3cLQUSD8gVq;`wW@l(=YHMt5Zf|gLa&vTbc6WGrdV73*et&?0f`f#GhKGoWii?bmj*pO$l9QB`mY0~Bnwy-Ro}ZwhqNAjxrl+W>s;jK6uCK5F00000001W32pkmz88rj|1{oLvX=D`)7UBXHv>F!F2qwq`CI}oD2(bVF0000002l}X000000000M04NBt000000000O2mt^900000000>*6$BLo6#@Vi02Krk0ssI26$BM<6-Z1K4iy9v001bpFa01m-2ecf0{{d71po#B2LK2F2><{90000022=t70|6BP6)Y7#EEVDc6$BLq6$Ts_2nhfH0000002l}Z000000000M04N9v000000000O2m=5B0000000000D77#BAQrIz00RI30OA4x0|Q7TNqho&12zO41qucxwhSB<1Q|6100tQt2ykp+Ze?^yWEBh+;sO@50v4VECj1H}m<k*i2nhfH0000002l}Z000000000M04N9v000000000O2m=5B00000000>*6$BLo6#@Vi02Krk0ssI26$BNy9u<xdOcmk+6$BCh04TLD{a~C1?H>S&K}-uO@QQ4618G93ChMWpP|8}Or78Fxgq<KB!tMWCK%PE{6O&D3<tM-tAqAoUuK=z9FRcm1>}qQgH}Zd>Rkc(r<?+=p{5V>XWxL6~@MCCz-}VIN)PT@jE~`$8K4mVOBx>eL%>RbQbyH>@9N|*hFx<Kk;6}AXn9Uq2X~Qs$_^(jS%NL<KCEDV`E284J(SL?!s}|;^(@p*_a$QnO`QJ!w(Ipl6@8Ag6xs?2OC{yZsS>V4TD{gD@pTP$H1At7eI{arqmbhL1Gaj?$zaslul@2;t$}Y!w;T9VQ+00&p`;sn3{x29-mrIUNonoFkrD%&vZ8fo8R#Z}Mm#9@f-qz}({CB*Ji%D$=cpd&P;MJmHES65EH~Ig7NiC$}KoImmP{ny52!R0$0w52BKtO;(0;J+3JRqpzJkJ9O5W!YT^Hb+8iL!L6u_jfgM6*sQ&23tTpl#}f^4pEmhGEJPN}uI1r5UAYiwpDO5`sOhqJ`XIAK2!q#jVnT|BPITO7OqobUPt(7#>LEFg%g7@I=nSg8_;hg#;*a6e36AL12g+g=a!1s?{#AN%%S|R&2{T)<hZJPKecLxcl0el>dsz3|{j?KYjSm*l4BH`S9P7E<$Zs%4wz_@9{QC8~!_bGrA2rCH^m(y1Kp2jjG)4NsFo3Z1YMk^53zY3*542S^gtTSleJb^Q;)H=)Yqkdl?vL&^eLovX@fuU-4Koqmf9PD@OknkXxlzh-TMjCH@~gZU!XbC_t51WlC{IWc2?)TTCk9WnEG%jocRqw=+Uxx-9Uvg6aTax@;_$t+q=1XFQJEY%5Fhe<7tME^EEyTh<!2Gt+6A;7%|5!CBMW#=LBmM#z#zjan<n?bu}VbVVtxX0k5-70KSZ27BugGJ=iGQP)|O$CNVoNVX7aPcOH@X{!$WZ{#{<%@X#q*Ajq|s39Drq_j+lh)APKb6}z|MpYgV7}GR`4p>86l{BE~gDC@sB=*5wiQqU!B@W&vBDB_7nf_Dh&n3`YpTmR82$DgByY^>)Kr&J5Pd^#_2m?&x0p-b*pGD)FGyOCHmPi1Kb*&K%WJ+x$DdufoS8iXS!mjx1Wap!!b_B!e`m?CWdk@2J5>*$fw|JH{#x~VZE0rMe*B(@{(bP2YG8-sQY%u<Mv>FzFoavKAG$5SB2fORn3wH9FIYjUP$;%pZ9f`BIw8AivrAhmgk~ytF8gC%^G9?WmM?@I7wPC<neEBAnQ8xLRX_6nOlERk^_J#?BEL&$?He`&V$T9(Xy1U0a|Li$e!jHWFamq9IC|FGAyf|L`I`ajj<2I`b6#x~20~M45EEVDc6$BLq6$TtO04M}mZE12C2tj0WVRLk4VE_OCD77#BAi)U+00RI30OA4x2rmEu0s{mE1_uZU3JVMk4i69!5)%{^78e*98XFuP9v>hfA|oUvCMPH<Dl054E-x@KGBY$aHa9pqIy*c)K0iP~LPJDFMn_0VN=r;lPESx#Qd3k_R##YAT3cLQUSD8gVq;`wW@l(=YHMt5Zf|gLa&vTbc6WGrdV73*et&?0f`f#GhKGoWii?bmj*pO$l9QB`mY0~Bnwy-Ro}ZwhqNAjxrl+W>s;jK6uCK5F00000001VGA{-SA88rj|1{oLxZeeX@6$}>Q0v4<y7L*hwupTCW4jeWBC<Iw;X>u3{L1c1ab97~402wS51Qi4o0ss{N6$BLm0000L1Qp;66{rhL70dw@1QGxMD77#BU@!tz765c783a}i%_g%zVxz3zOiV{Vc59go$MYFKBis3Y>PNRgeLpo`dIhOKZxVyR>QBQ!i)`onNejF{M2l?a`>DY{&CJZq%*@<B83a}iEwZgYdlarm;d+#GKNXBh>I5`#KQC^h0lWo*Km4ij(oJGuKeA&Z5U=Y;Kbo1DnVFfnK!87xzDd)dXbwMng$4Hm9tc0TUp`qeT<t!hCi;Ao5$8VfXyb^r_|!g)?X{ry#^62$q9<iiL+d_N<PdRlD#<?n*|0tUzQGqJQw0P-$e6p9kk1@IrV2(Sb(%y#$Ok=}RYy`lwWfK;DL!LBC!mS25sAS-UkA<^QQFr(WA0i)J|JE|-L-^#(6iw`?G3&#nJU0P4d5-1p!lIbo(ztXSLaASi{O7N*|9M{fS+;UKu9J(ORzYc`PDo>kb@Ft+Wv(<u``9vYwUtQID{%Sd!RKxz&=7h57z%V%$FxW6f8f-y10!7@K!&`$L<0p1@=Cf@*$`iy$U~+_6A>=Ol9~!RX|`=KxSrUW@d(LKn!R=5kUHWYP@ulXg`w6e@MT5Z9fxX!_zBec|R9OQ2ZEkmp^nyl=cQ+h(FGV(%#?;qd%H=oT4MSyg!`8Ah2?1k(fW$#ceczx5z&V*Q0Pf3dcW2C3OOt2zx)^oDrqH!FNAfKdDWe-0x5{Khpcg*PdBIKfHOWLvDsUKVdReFe<52Lq9D;9P8pXDnCw^Kmb5Fw8(b8pNg43>`}NLh3mmT0MI`lh3iqc9@RhLdK9il;R=XrKRe$~jhAk6Ki+YQj^y%hKe|Z_0xO4xGe0vkGcz-jSwK}kdQv|f0K+7UIR>8Ef!ZUFiDhAI8P=xt9m1U{f@5bI%E~<rE&Kt<X6U%ty$QG0widU{Y$iiwv9^I$c3{ZJ42jFLRKA(agaMCO$t*jwzLu3J6#x~m0u{mnEEVDc6$BLq6$Ts_2yYUD!0J!KKo|%B00000004j%04NA=5`)0%Ps2bM2mk;800000fB*mhD77#BAeg}f00RI30OA4x2OWTbU_dAs4v2(e!Du)hkcebLsaP(UOh$9ra5x-|M$_qZIvq`i)6sM|8cwIv>2x|Bj;G`4bUK|*r_<?tKq1kHR5G1VDb<SAa=l=knCIg0z+^L8rB<=saJh=k<MDVrCLRqtUeEUf28Ban(Rf5A70M;E>3l+=QR$RgwO+AV?Uvj1e!*e!m|XS%04B0B92E)~H3R?#85ja*Z50d_;sO@35EhINCh#sMrXw5}2yYUD!0J!KKo|%B00000004j%04NA=5`)0%Ps2bM2mk;800000fB+dR6$BLo6#@Vi02Krk0ssI26$BOV3l;bTOclNX6$BCh04TLD{b2Y3l?ec(3;+NC001VNKr<#l&CD1;W@ct)W_Fxon2w}CnwgoInbe}(R6qa#0000$Re(hRJ~J{mEF`8#W*~qyGZ6p~nFR#?KLN}H`9CuYKmY*vKk(o`#%P0YsDO=u;)g&fT#v$a9i-r3{Dz=fU{(xGnK8^LW+gK&j73>QEX?#jW)AGrKO!bT9zYljATSg(2p57yf~Sbfhk0!oG&MVbL_tbX078I9Kvcj|0691+C?%CZWq?2cRzO&406#W0+=4&=00000fIdK3s1R6NHVyzIve^L;5Q9ET)B2l>Y4+iNXMXBl8f=&1w$3&$q&@k6zMFO7sW#kJLhQ7_-Z~aH9C-cRnAxzA>-;l%S%{~{oX{@hL@x8hxqVLC%Z|Y|^m{wMoK_r~-6hyg$Q1w;)B+Xe0xT8c0u=-m1{DSz7zk!&W@cuNq(B%5Gcz+YGc(ga762#+W@ct)W{#vl7zi^nGcz+Y(?0+J04TLD{UEAQ1pos80080w0S7ODfM7r<7!HVpV!>!Q9*~G+LaA6Tm`p}<*>E_VPN(DXcs!nthvVsZI3141<MDVr9?r+}@pwEQkH_QjfI^}Xsbo5#QmPfJ<$A$l8jEa3tJ!Y2oNlM%QF=a|?)UrYES%r(_xt^Rzu)g?`Eh=h9|sr|4v9tM5t&pjnN272361h>x5q;|wNa~9<nxNnYJFR-!q#ng9S;Bi00000Ch$BQ6$%+O1ONsZ7zT1-WNCGC6$}>Q0v6^E7JvyRia93mGaMKQW@ct)W{#vl7zi^nGcz+Y(?1pfC<ta|W@cuNq(B%5Gcz+YGc(ga02wS51Qi4o0ss{N6$BLm0000L1Qmb<6~qKg6;Krf5&!@wwJ-f(Kmf%E05lJcmu?b+zzRUY%ArNJ^Zh^Yr^ZV+iT6L#O=1vOISfFw$acP;8VEq&PmPyu68}H)(oJFzSp7d<x=9QIE9*ba%*@Qp%mhFH2ta1?Kkj@#HD0>;KZ!wL<<KJUKLl0|EwY{W^FQe(F$k<2{y&_=Ah2?1;Xj8K+0OS<@jqyh?R-Bq`adaLkHYmR<Ue~9u1Dc|<Uau5KXE+@*Q0RVKe!%+>ruGgKV08G;d&IVN8#E(!}Ta!kHXbIW&%J)w)6eec=`7~WINwajhFgAi)`onsqy$f27#4Bi){1sKlncYKsovVK$@AEnVFgC_CE~(W5_o&FTrYKD~w^_mX^oPq=aj?Qn4b2yW5!<G?xe<3KakqrU4bW0W1~b0u=-m1{DSz7zoo%Vh~t43_utN!}Ta!kHXbI76K>;(@kO!SUC(p7zo4lC|r-i)jt3L04TLD{UB4B0ssX7007_u0RjL913(}QL}45d2m?uyAP9mWTnGos!ZgkUMbbo7Rb^dRR#lZ&RZ1HLT<3jY7=~dOhG7_r5F!U*7>;5XhG7_rq9}?XS(+yR04CB!92E)~H3R?#85jp-Wo~n6ba@pF7UBXH#s?Oh2PT?BCbm8t7zoo%Vh~t43_utN!}Ta!kHXbI76K>;(@kO!SUC(p7zo4lC|r-i)jt3kEENP51Qh}R6#x|k6#@VN02KrkunZOE1xyvj0Tl!i001bpFa2N`0!<45q75}PaBJEyy1W$E2-=V0H^?B?Jx4xx=)@QG-L0OdKhRZyUN?GW?<8p(5G^7xgD6<4jl&gWpWrBQODNPvU@*MQfILcR#~-roPTKoL1+)1?k2Ty$%QW66?|6@Q(z5NoRM1?gki|<1E!*zEUlIr$APeVZ-r9iFxhb3$Q#bRJD%_ZrF6gdGPyiqLQ4gx=cB~R(nXQyE#|7sRM(f+NJ4|9H8A_9g_I$*>RW0}e7-Nhl9sslXL=}e*ph6XzOZD$Pwd)&bt+n=aDhQB;^D+TfsIh2Q%c)atPy$gZiyIISWAOt+!Q0rl+e=y_cTdant7KEOsog{KxQL6plyb-V??~H_Cs@F>CUoB5lQ1=6qi6WYn6ymL<A8C#0I#qy9HJd0ZHO%z)3(QX$dzSVM_4owyJutVnY4NFoxk392upC?(v+=_kMq*UrmchPXLb6JkN`aB^rZm3(mFJ_R22Xf&;k|W0xT8c0u=-m1{DSz7zh9W00000J3tr+dlarm;d&{+77Qo|000000028c7zle5u1Dc|DZl^#04TLD{UE481po*D006K70RzCI0RaasfPi2?C>Rcigkr&9G#rTsBqEtmDwfLyQ^{yHoK6Sh@q9iYOsFElj7T36sAM{UOrO)Jgc5~JqY@}J5~)_Kmg@zJ$zU^D?P0sgaFkq!gyQISB@+*&>-l~<U{Kf-p@c&su!kBFhee^VXFMX4N9B^)bUvX`>6ChkfI_}%uxKp_uUJP`tJ!j?!)~`L^nOj?uowUU001V0Q5+Qt88rj|1{oL$ZDD6_X>@gDWfcq-;sO?;5EkkQCZbIynnxTM2mk;80000xKo|&n6s||%dMUsb3@8Wy0000006Rb!2zwN+N8x%YzyKL66$BLo6#@Vi02Krk0ssI26$BNa1{JCWOci7m1QGxMD77#BV2}V^3IJpgxE_V;QMkT8lNbb64lS}joWvloa%izXDO``j^(f>&aXkvxqj2Rvh3iqc9);UK!}Ta!kHW=23fH4>Jqou!dlarm;d-P$s`1iIVi2f5nwgoInVGpi2&^1hWIL}vVh~t4w8-v10QNsbi)`onsq;Tp4lS~s@9#g6?R-BqUhc_1Gcz+YGc(iUKO@`uermk5KjuGN=lbyV+2F1}_9$GB!u8lc;d&IVN8!ppNYTyIKkQMs9);_%KcsLy3fH5su|FFx-6RHqwX;7v-%pK~Zn8i0{nU8rCb2&`w8(b8pSrp~HD0<&3<Bpr*TX-uo$sf{OSeDMO+Wg6YP@ulxIY3bhZfn+*FR=vW@ct));|*f6)Rt5FKTLeY%I$&%V24;Y?U`l3rjiKakrpjt9P;mcV`^dyud3n&`LgSl_PCI6cqp!xB(T)0W1~b0u=-m1{DSz7znr?h3iqc`ac*5dlarm;d-P$77Qo|xE_V;QMmd)7zle5u1Dc|q(1-v04TLD{UBW70ssgA006K70RzCI0RaXBKp+f6VH^+$1BtRA2!bG72nkH%JWwP}RApUQrcK+%aU92S9LI4ShafHp1)cYSp(u`IX_6<JA`l1%qAITJ*&c5DA_#&Y3LM9A3;-tNS{xM$88rj|1{oLzVQgY$WN#G=7UBXH_y`u12qv0XCeTtG7znr?h3iqc`ac*5dlarm;d-P$77Qo|xE_V;QMmd)7zle5u1Dc|q(1-|EENP51Qh}R6#x|k6#@VN02KlmoHYO#26JO*Wo=;<5C9bfB?1@%X=DHu1SJ9(2ykp+Ze?^yWB?TmB?1@(ZeeX@B>+qu0000L3MB#<0%vUi6$&K+7zT1-WNCGC02K-)0vHEmWo~n6ba?<33MB#<32k9#ZfSINWMu#q3MB#<24QSsWn^yv7UBXK92txz+z1>M1Q|6100tQt0%>Fw3>M-77PJ}`)CeZX1SSX^7znWd00000000;W0RR9100000762#+u>b%7000007zhCX000000000PEENP51Qh}R6#x|k6#@VN04BB!92Eo^H3R?#85js~Y+-I?bV+0t3>M-77PJBuo&qNP3MQBe92f`*000000000O2m=5B0000002Tl!2nhfH0000002l}Z000000000087vh96$BLm02Kfg1Qh}R001VGA{-SA88rj|1{oLxZeeX@6$}>Q0v4<y7L*hwupTCW4jeWBC<Iw;X>u3{L1c1ab97~402wS51Qi4o0ss{N6$BLm0000cvN9YM3K=y700tQt0%vU%3>M-77P1f)j1MO8E+(cU92f|15`)0%Ps2bM2mk;800000fEEBK2yYUD!0J!KKo|%B00000004jh87vh96$BLm02Kfg1Qh}R001WNJRB7Y88rj|1{oLza$#g?b#oOA7UBXH<_{Kt2_}j;Ch#*H7zk!&W@cuNq(B%5Gcz+YGc(ga762#+W@ct)W{#vl7zi^nGcz+Y(?0+iEENP51Qh}R6#x|k6#@VN04CB!92E)~H3R?#85jp-Wo~n6ba@pF7UBXH#s?Oh2PT?BCbm8t7zoo%Vh~t43_utN!}Ta!kHXbI76K>;(@kO!SUC(p7zo4lC|r-i)jt3kEENP51Qh}R6#x|k6#@VN049V{92E)~H3R?#85jv|VP|e>baiB96$}>Q0v4hW7U~HmqD>~6M;sUk000000028c7zle5u1Dc|DZmyCC<p)m00000J3tr+dlarm;d&{+02wS51Qi4o0ss{N6$BLm0000c<XRjR3K=y700tQt24QSsWn^y^3>M-77WfDjln5r8SSHX?92f|=9);^sxcWaB2zwN+N8x&;KNbuq2)G`F>ruG+KNtvm6s||%dZa%987vh96$BLm02Kfg1Qh}R000(@fB_cb0w&xD7OYSd000>r7z{yDQcqVpb7N>_ZDAOI1urizFE2S~LP0@6Q9(gLK|w)5OF=<*K|*IiMnOSmK|^OjK|w)5LP9}7O+i6HK|?`7NkKtDK|(=6NkKtDK|(=6K|w)hK|w)6FhNm4K}<nKK|xS=K|w)8NkKtDL2W@nK|x1nK|w)5FhM~<K|(=6K|xG<D=#lNK|w)9L`6YCK|w-iK|w)5ML|J9K|w)5K|w)TK|w)6Xktf7Y+6S+K|w_~FE1}RK|w)DK{ik~FE4jNK|xVXQ9(gLK}bPCK|w)LK|w)5K|w)5K|@hNK|x0`Xj(EgaAZhUYg$o3K|x42FE1}RK|w)DK}S$8FE4jNK|xVXQ9(gLK}SJBK|w)LK|w)5K|w)5K|*&yK|x1SY+`OsaAZhpL2-63FE4jNK|yCiMK3QeFF`>;LP2;zK|w)5SwTTTK|(=6K|w)5K|w)5M?pbBL1S%3YhhScdO<-%S1&IwFF`>;LqSMHFE1}RK|w)9L`6YCK|w-yK|w)5ML|J9K|w)5K|w)DK|w)6ZemeEK|w)5M?pbnK|*IiM?pbBK|w)5LqSVKFE1}RK|w)9MnyqEK|w-yK|w)5ML|J9K|w)5K|w)LK|w)6by-(2Y(YUmK|w)LK|(=6ML|J9L3&UxFE4jNK|xVTQ9(gLK}bPCK|w)LK|w)5K|w)5K|?`7K|x4Db6Hn4Y-B}jYe7LlK|xJaFE1}RK|w)5LQz3MK|w-6K}JDANkKtbK|*&yO+i6HK}A79Q9(gLK|w)5LP2OjK|w)5V?jYdK|(=6K|w)5K|w)5LqS17L19->K|w)hK}118NkKtlK|?`7K|w)5K|w)AK|w)5K>!#ZaA9(Bb!BudV{mXSVRCYBcOZ6Ua&u{KZXh)-FfK3|j2r*}8~^|u000~S02}}S8~^|u000~S003PE002-yQZW"
            )
        )
    )


def schema():
    # fmt:off
    return RelationSchema(
            name="$satellites",
            columns=[
                FlatColumn(name="id", type=OrsoTypes.INTEGER),
                FlatColumn(name="planetId", type=OrsoTypes.INTEGER, aliases=["planet_id"]),
                FlatColumn(name="name", type=OrsoTypes.VARCHAR),
                FlatColumn(name="gm", type=OrsoTypes.DOUBLE),
                FlatColumn(name="radius", type=OrsoTypes.DOUBLE),
                FlatColumn(name="density", type=OrsoTypes.DOUBLE),
                FlatColumn(name="magnitude", type=OrsoTypes.DOUBLE),
                FlatColumn(name="albedo", type=OrsoTypes.DOUBLE),
            ],
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Explain Node

This is the SQL Query Execution Plan Node responsible for the reading of data.

It wraps different internal readers (e.g. GCP Blob reader, SQL Reader), 
normalizes the data into the format for internal processing. 
"""
import time
from typing import Generator

import pyarrow
from orso.schema import RelationSchema

from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


def normalize_morsel(schema: RelationSchema, morsel: pyarrow.Table) -> pyarrow.Table:
    if len(schema.columns) == 0:
        one_column = pyarrow.array([1] * morsel.num_rows, type=pyarrow.int8())
        morsel = morsel.append_column("*", one_column)
        return morsel.select(["*"])

    # rename columns for internal use
    target_column_names = []
    # columns in the data but not in the schema, droppable
    droppable_columns = []

    # find which columns to drop and which columns we already have
    for i, column in enumerate(morsel.column_names):
        column_name = schema.find_column(column)
        if column_name is None:
            droppable_columns.append(i)
        else:
            target_column_names.append(str(column_name))

    # remove from the end otherwise we'll remove the wrong columns after the first one
    droppable_columns.reverse()
    for droppable in droppable_columns:
        morsel = morsel.remove_column(droppable)

    # remane columns to the internal names (identities)
    morsel = morsel.rename_columns(target_column_names)

    # add columns we don't have
    for column in schema.columns:
        if column.identity not in target_column_names:
            null_column = pyarrow.array([None] * morsel.num_rows)
            morsel = morsel.append_column(column.identity, null_column)

    # ensure the columns are in the right order
    return morsel.select([col.identity for col in schema.columns])


class ScannerNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **parameters):
        super().__init__(properties=properties, **parameters)
        self.start_date = parameters.get("start_date")
        self.end_date = parameters.get("end_date")
        self.hints = parameters.get("hints", [])
        self.columns = parameters.get("columns", [])
        self.predicates = parameters.get("predicates", [])

        if len(self.hints) != 0:
            self.statistics.add_message("All HINTS are currently ignored")

    @property
    def name(self):  # pragma: no cover
        """friendly name for this step"""
        return "Scan"

    @property  # pragma: no cover
    def config(self):
        """Additional details for this step"""
        date_range = ""
        if self.parameters.get("start_date") == self.parameters.get("end_date"):
            if self.parameters.get("start_date") is not None:
                date_range = f" FOR '{self.parameters.get('start_date')}'"
        else:
            date_range = (
                f" FOR '{self.parameters.get('start_date')}' TO '{self.parameters.get('end_date')}'"
            )
        return (
            f"({self.parameters.get('relation')}"
            f"{' AS ' + self.parameters.get('alias') if self.parameters.get('alias') else ''}"
            f"{date_range}"
            f"{' WITH(' + ','.join(self.parameters.get('hints')) + ')' if self.parameters.get('hints') else ''})"
        )

    def execute(self) -> Generator:
        """Perform this step, time how long is spent doing work"""
        morsel = None
        orso_schema = self.parameters["schema"]
        orso_schema_cols = []
        for col in orso_schema.columns:
            if col.identity in [c.identity for c in self.columns]:
                orso_schema_cols.append(col)
        orso_schema.columns = orso_schema_cols
        arrow_schema = None
        start_clock = time.monotonic_ns()
        reader = self.parameters["connector"].read_dataset(
            columns=self.columns, predicates=self.predicates
        )
        for morsel in reader:
            morsel = normalize_morsel(orso_schema, morsel)
            if arrow_schema:
                morsel = morsel.cast(arrow_schema)
            else:
                arrow_schema = morsel.schema
            self.statistics.time_reading_blobs += time.monotonic_ns() - start_clock
            self.statistics.blobs_read += 1
            self.statistics.rows_read += morsel.num_rows
            self.statistics.bytes_processed += morsel.nbytes
            yield morsel
            start_clock = time.monotonic_ns()
        if morsel:
            self.statistics.columns_read += morsel.num_columns



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Tuple

from orso.tools import random_int

from opteryx.shared.variables import SystemVariables
from opteryx.shared.variables import SystemVariablesContainer
from opteryx.shared.variables import VariableOwner

# History Item = [statement, success, execution start]
HistoryItem = Tuple[str, bool, datetime.datetime]


@dataclass
class ConnectionContext:
    """
    Manages the context for each database connection.

    Attributes:
        connection_id: int
            Unique identifier for the connection.
        connected_at: datetime.datetime
            Timestamp indicating when the connection was established.
        user: str, optional
            Username for the connection, defaults to None.
        schema: str, optional
            Schema to be used in the connection, defaults to None.
        variables: dict
            System variables available during the connection.
        history: List[HistoryItem]
            A history of queries executed in this connection.
    """

    connection_id: int = field(default_factory=random_int, init=False)
    connected_at: datetime.datetime = field(default_factory=datetime.datetime.utcnow, init=False)
    user: str = None
    schema: str = None
    variables: SystemVariablesContainer = field(init=False)
    history: List[HistoryItem] = field(default_factory=list, init=False)

    def __post_init__(self):
        """
        Initializes additional attributes after the object has been created.
        """
        # The initializer is a function rather than an empty constructor so we init here
        object.__setattr__(self, "variables", SystemVariables.snapshot(VariableOwner.USER))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def suggest_alternative(name, candidates):
    """
    Find closest match using a variation of Levenshtein Distance

    This implementation is limited to searching for distance less than three, is case
    insenstive and removes any non-alpha numeric characters.

    This is tuned for this use case of quickly identifying likely matches when a user
    is entering field or function names and may have minor typos, casing or punctuation
    mismatches with the source value.
    """
    from opteryx.third_party.mbleven import compare

    best_match_column = None
    best_match_score = 100

    name = "".join(char for char in name if char.isalnum())
    for raw, candidate in (
        (
            ca,
            "".join(ch for ch in ca if ch.isalnum()),
        )
        for ca in candidates
    ):
        my_dist = compare(candidate, name)
        if my_dist == 0:  # if we find an exact match, return that
            return raw
        if 0 <= my_dist < best_match_score:
            best_match_score = my_dist
            best_match_column = raw

    return best_match_column



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
no table
---------

This is used to prepresent no table.

It actually is a table, with one row and one column.
"""
from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import OrsoTypes


def read(*args):
    import pyarrow

    # Create a PyArrow table with one column and one row

    return pyarrow.Table.from_arrays([[0]], ["$column"])  # schema=_schema)


def schema():
    # fmt:off
    return RelationSchema(name="$no_table", columns=[FlatColumn(name="$column", type=OrsoTypes.INTEGER)])
    # fmt:on



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Selection Node

This is a SQL Query Execution Plan Node.

This node is responsible for applying filters to datasets.
"""
import time
from typing import Generator

import numpy
import pyarrow

from opteryx.exceptions import SqlError
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import evaluate
from opteryx.managers.expression import evaluate_and_append
from opteryx.managers.expression import format_expression
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


class SelectionNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self.filter = config.get("filter")

        self.function_evaluations = get_all_nodes_of_type(
            self.filter,
            select_nodes=(NodeType.FUNCTION,),
        )

    @property
    def config(self):  # pragma: no cover
        return format_expression(self.filter)

    @property
    def name(self):  # pragma: no cover
        return "Selection"

    def execute(self) -> Generator:
        morsels = self._producers[0]  # type:ignore
        schema = None
        at_least_one = False

        for morsel in morsels.execute():
            if schema is None:
                schema = morsel.schema

            if morsel.num_rows == 0:
                continue

            start_selection = time.time_ns()
            morsel = evaluate_and_append(self.function_evaluations, morsel)
            mask = evaluate(self.filter, morsel)
            self.statistics.time_evaluating += time.time_ns() - start_selection

            if not isinstance(mask, pyarrow.lib.BooleanArray):
                try:
                    mask = pyarrow.array(mask, type=pyarrow.bool_())
                except Exception as err:
                    raise SqlError(
                        f"Unable to filter on expression '{format_expression(self.filter)}'."
                    )
            mask = numpy.nonzero(mask)[0]

            self.statistics.time_selecting += time.time_ns() - start_selection

            # if there's no matching rows, just drop the morsel
            if mask.size > 0 and not numpy.all(mask is None):
                yield morsel.take(pyarrow.array(mask))
                at_least_one = True

        # we need to send something to the next operator, send an empty table
        if not at_least_one:
            yield pyarrow.Table.from_arrays([[] for i in schema.names], schema=schema)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Bespoke error types for Opteryx and error types and structure as defined in PEP-0249.

Exception Hierarchy:

Exception
 ├── MissingDependencyError
 ├── UnmetRequirementError
 └── Error [PEP-0249] *
     └── DatabaseError [PEP-0249] *
         ├── IncompleteImplementationError
         ├── InvalidConfigurationError
         ├── InvalidInternalStateError
         ├── NotSupportedError
         ├── UnsupportedFileTypeError
         ├── UnsupportedTypeError
         └── ProgrammingError [PEP-0249] *
             ├── DataError *
             │   ├── InconsistentSchemaError
             │   ├── EmptyDatasetError
             │   └── EmptyResultSetError
             ├── ExecutionError *
             ├── MissingSqlStatement
             ├── InvalidCursorStateError
             ├── ParameterError
             ├── SecurityError *
             │   └── PermissionsError
             └── SqlError *
                 ├── AmbiguousDatasetError
                 ├── AmbiguousIdentifierError
                 ├── ArrayWithMixedTypesError
                 ├── ColumnNotFoundError
                 ├── ColumnReferencedBeforeEvaluationError
                 ├── DatasetNotFoundError
                 ├── FunctionNotFoundError
                 ├── IncorrectTypeError
                 ├── InvalidFunctionParameterError
                 ├── InvalidTemporalRangeFilterError
                 ├── IncompatibleTypesError
                 ├── UnexpectedDatasetReferenceError
                 ├── UnnamedColumnError
                 ├── UnnamedSubqueryError
                 ├── UnsupportedSyntaxError
                 └── VariableNotFoundError
"""

from typing import Any
from typing import Optional


# ======================== Begin Codebase Errors ========================
class MissingDependencyError(Exception):
    def __init__(self, dependency: str):
        self.dependency = dependency
        message = f"No module named '{dependency}' can be found, please install or include in requirements.txt"
        super().__init__(message)


# ======================== End Codebase Errors ==========================


# ======================== Begin PEP-0249 Exceptions ========================
# These should not be thrown directly unless explicitly required for standards compliance
class Error(Exception):
    """
    https://www.python.org/dev/peps/pep-0249/
    Exception that is the base class of all other error exceptions. You can use this to
    catch all errors with one single except statement. Warnings are not considered
    errors and thus should not use this class as base. It must be a subclass of the
    Python StandardError (defined in the module exceptions).
    """


class DatabaseError(Error):
    """
    https://www.python.org/dev/peps/pep-0249/
    Exception raised for errors that are related to the database. It must be a subclass
    of Error.
    """


class ProgrammingError(DatabaseError):
    """
    https://www.python.org/dev/peps/pep-0249/
    Exception raised for programming errors, e.g. table not found or already exists,
    syntax error in the SQL statement, wrong number of parameters specified, etc. It
    must be a subclass of DatabaseError.
    """


# ======================== End PEP-0249 Exceptions ==========================


# ======================== Begin Opteryx Superclasses ========================
# These should not be thrown directly
class SqlError(ProgrammingError):
    """
    Used as a superclass for errors users can resolve by updating the SQL statement.

    Where possible, SqlErrors in particular, should provide messages appropriate for
    end-users who may not know, or care, about the underlying SQL platform.
    """


class DataError(ProgrammingError):
    """Superclass for data-related errors."""


class SecurityError(ProgrammingError):
    """Superclass for security-related errors."""


class ExecutionError(ProgrammingError):
    """Superclass for execution-related errors."""


# ======================== End Opteryx Superclasses ==========================


# ======================== Begin SQL-Specific Exceptions ========================
class ColumnNotFoundError(SqlError):
    """Exception raised for Column Not Found errors."""

    def __init__(
        self, message: str = None, column: str = None, dataset: str = None, suggestion: str = None
    ):
        """
        Return as helpful Column Not Found error as we can by being specific and offering
        suggestions.
        """
        self.column = column
        self.suggestion = suggestion
        self.dataset = dataset

        dataset_message = (f" in '{dataset}'") if dataset else ""
        if column is not None:
            message = f"Column '{column}' does not exist{dataset_message}."
            if suggestion is not None:
                message += f" Did you mean '{suggestion}'?."
        if message is None:
            message = "Query contained columns which could not be found."
        super().__init__(message)


class ColumnReferencedBeforeEvaluationError(SqlError):
    """
    Return an error message when the column reference order is incorrect
    """

    def __init__(self, column: str):
        self.column = column
        message = f"Reference to '{column}' cannot be made here, it hasn't been evaluated yet due to the internal order of query evaluation."
        super().__init__(message)


class DatasetNotFoundError(SqlError):
    """Exception raised when a dataset is not found."""

    def __init__(self, dataset: str = None, suggestion: Optional[str] = None):
        self.dataset = dataset
        message = f"The requested dataset, '{dataset}', could not be found."
        if suggestion is not None:
            message += f" Did you mean '{suggestion}'?"
        super().__init__(message)


class FunctionNotFoundError(SqlError):
    """Exception raised when a function is not found."""

    def __init__(self, message: str = None, function: str = None, suggestion: Optional[str] = None):
        """
        Return as helpful Function Not Found error as we can by being specific and offering
        suggestions.
        """
        self.function = function
        self.suggestion = suggestion

        if message is None:
            message = f"Function '{function}' does not exist."
            if suggestion is not None:
                message += f" Did you mean '{suggestion}'?."
        super().__init__(message)


class VariableNotFoundError(SqlError):
    """Exception raised when a variable is not found."""

    def __init__(self, variable: str, suggestion: Optional[str] = None):
        if variable is not None:
            self.variable = variable

            message = f"Variable '{variable}' does not exist."
            if suggestion is not None:
                message += f" Did you mean '{suggestion}'?"

            super().__init__(message)
        else:
            super().__init__()


class AmbiguousIdentifierError(SqlError):
    """Exception raised for ambiguous identifier references."""

    def __init__(self, identifier: Optional[str] = None, message: Optional[str] = None):
        self.identifier = identifier
        if message is None:
            message = f"Identifier reference '{identifier}' is ambiguous; Try adding the databaset name as a prefix e.g. 'dataset.{identifier}'."
        super().__init__(message)


class AmbiguousDatasetError(SqlError):
    """Exception raised for ambiguous dataset references."""

    def __init__(self, dataset: str):
        self.dataset = dataset
        message = f"Dataset reference '{dataset}' is ambiguous; Datasets referenced multiple times in the same query must be aliased."
        super().__init__(message)


class UnexpectedDatasetReferenceError(SqlError):
    """Exception raised for unexpected dataset references."""

    def __init__(self, dataset: str):
        self.dataset = dataset
        message = f"Dataset '{dataset}' referenced in query without being referenced in a FROM or JOIN clause."
        super().__init__(message)


class InvalidTemporalRangeFilterError(SqlError):
    """Exception raised for invalid temporal range filters."""


class InvalidFunctionParameterError(SqlError):
    """Exception raised for invalid function parameters."""


class UnsupportedSyntaxError(SqlError):
    """Exception raised for unsupported syntax."""


class IncorrectTypeError(SqlError):
    """Exception raised for incorrect types."""


class IncompatibleTypesError(Exception):
    """
    Raised when attempting to join fields of incompatible types.

    Parameters:
        left_type: str
            The type of the left field.
        right_type: str
            The type of the right field.
        column: Optional[str]
            If the incompatibility occurs in a single column
        left_column: Optional[str]
            The column name where the error occurs.
        right_columns: Optional[str]
            The column name where the error occurs.

    Attributes:
        left_type (str): The type of the left field.
        right_type (str): The type of the right field.
        column (str): The column name where the error occurs.
        left_column (str): The column name where the error occurs.
        right_column (str): The column name where the error occurs.
    """

    def __init__(
        self,
        left_type: str,
        right_type: str,
        column: Optional[str] = None,
        left_column: Optional[str] = None,
        right_column: Optional[str] = None,
        left_node: Optional[Any] = None,
        right_node: Optional[Any] = None,
    ):
        def _format_col(_type, _node, _name):
            if _node.node_type == 42:
                return f"literal '{_node.value}' ({_type})"
            if _node.node_type == 38:
                return f"column '{_name}' ({_type})"
            return _name

        self.left_type = left_type
        self.right_type = right_type
        self.column = column
        self.left_column = left_column
        self.right_column = right_column
        if self.column:
            super().__init__(
                f"Incompatible types for column '{column}': {left_type} and {right_type}"
            )
        elif self.left_column or self.right_column:
            super().__init__(
                f"Incompatible types for {_format_col(left_type, left_node, left_column)} and {_format_col(right_type, right_node, right_column)}. Using `CAST(column AS type)` may help resolve."
            )
        else:
            super().__init__("Incompatible column types.")


class ArrayWithMixedTypesError(SqlError):
    """Exception raised when arrays have mixed types."""


class PermissionsError(SecurityError):
    """Exception raised for permissions errors."""


# ======================== End SQL-Specific Exceptions ==========================


# ======================== Begin Miscellaneous Database Errors ========================
class UnsupportedTypeError(DatabaseError):
    """Exception raised when an unsupported type is encountered."""


class UnmetRequirementError(Exception):
    """Exception raised when a requirement for operation is not met."""


class NotSupportedError(DatabaseError):
    """Exception raised when an unsupported operation is attempted."""


class UnsupportedFileTypeError(DatabaseError):
    """Exception raised when an unsupported file type is encountered."""


class MissingSqlStatement(ProgrammingError):
    """Exception raised for missing SQL statement."""


class InconsistentSchemaError(DataError):
    """Raised whem, despite efforts, we can't get a consistent schema."""


class EmptyDatasetError(DataError):
    """Exception raised when a dataset is empty."""

    def __init__(self, dataset: str):
        self.dataset = dataset
        message = f"The requested dataset, '{dataset}', appears to exist but no matching partitions were found."
        super().__init__(message)


class UnnamedSubqueryError(SqlError):
    """Exception raised for unnamed subqueries."""


class UnnamedColumnError(SqlError):
    """Exception raised for unnamed columns."""


# ======================== End Miscellaneous Database Errors ==========================


# ======================== Begin Configuration & Internal Errors ========================
class InvalidConfigurationError(DatabaseError):
    """Exception raised for invalid configuration."""

    def __init__(
        self, *, config_item: str, provided_value: str, valid_value_description: str = None
    ):
        DISPLAY_LIMIT: int = 32

        self.config_item = config_item
        self.provided_value = provided_value
        self.valid_value_description = valid_value_description

        message = f"Value of '{str(provided_value)[:DISPLAY_LIMIT]}{'...' if len(provided_value) > DISPLAY_LIMIT else ''}' for '{config_item}' is not valid."
        if valid_value_description:
            message += f" Value should be {valid_value_description}"
        super().__init__(message)


class InvalidInternalStateError(DatabaseError):
    """Exception raised for invalid internal states."""


class InvalidCursorStateError(ProgrammingError):
    """Exception raised for invalid cursor states."""


class ParameterError(ProgrammingError):
    """Exception raised for parameter errors."""


# ======================== End Configuration & Internal Errors ==========================



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
~~~
                      ┌───────────┐
                      │   USER    │
         ┌────────────┤           ◄────────────┐
         │SQL         └───────────┘            │
  ───────┼─────────────────────────────────────┼──────
         │                                     │
   ┌─────▼─────┐                               │
   │ SQL       │                               │
   │  Rewriter │                               │
   └─────┬─────┘                               │
         │SQL                                  │Results
   ┌─────▼─────┐                         ┌─────┴─────┐
   │           │                         │           │
   │ Parser    │                         │ Executor  │
   └─────┬─────┘                         └─────▲─────┘
         │AST                                  │Plan
   ┌─────▼─────┐                         ╔═══════════╗
   │ AST       │                         ║Cost-Based ║
   │ Rewriter  │                         ║ Optimizer ║
   └─────┬─────┘                         ╚═════▲═════╝
         │AST                                  │Plan
   ┌─────▼─────┐      ┌───────────┐      ┌─────┴─────┐
   │ Logical   │ Plan │ Heuristic │ Plan │           │
   │   Planner ├──────► Optimizer ├──────► Binder    │
   └───────────┘      └───────────┘      └─────▲─────┘
                                               │Schemas
                                         ┌─────┴─────┐
                                         │           │
                                         │ Catalogue │
                                         └───────────┘
~~~
This is written as a Visitor, unlike the binder which is working from the scanners up to
the projection, this starts at the projection and works toward the scanners. This works well because
the main activity we're doing is splitting nodes, individual node rewrites, and push downs.
"""


from opteryx.components.cost_based_optimizer.strategies import *
from opteryx.components.logical_planner import LogicalPlan

from .strategies.optimization_strategy import CostBasedOptimizerContext

__all__ = "do_cost_based_optimizer"


class CostBasedOptimizerVisitor:
    def __init__(self):
        self.strategies = [
            ConstantFoldingStrategy(),
            PredicatePushdownStrategy(),
            ProjectionPushdownStrategy(),
        ]

    def traverse(self, plan: LogicalPlan, strategy) -> LogicalPlan:
        """
        Traverse the logical plan tree and apply optimizations.

        Args:
            tree: The logical plan tree to optimize.

        Returns:
            The optimized logical plan tree.
        """
        root_nid = plan.get_exit_points().pop()
        context = CostBasedOptimizerContext(plan)

        def _inner(nid, parent_nid, context):
            node = context.pre_optimized_tree[nid]
            context.node_id = nid
            context.parent_nid = parent_nid
            context = strategy.visit(node, context)

            for child, _, _ in plan.ingoing_edges(nid):
                _inner(child, nid, context)

        _inner(root_nid, None, context)
        optimized_plan = strategy.complete(context.optimized_plan, context)
        return optimized_plan

    def optimize(self, plan: LogicalPlan) -> LogicalPlan:
        current_plan = plan
        for strategy in self.strategies:
            current_plan = self.traverse(current_plan, strategy)
        # DEBUG: log ("AFTER COST OPTIMIZATION")
        # DEBUG: log (current_plan.draw())
        return current_plan


def do_cost_based_optimizer(plan: LogicalPlan) -> LogicalPlan:
    optimizer = CostBasedOptimizerVisitor()
    return optimizer.optimize(plan)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Date Utilities
"""
import datetime
import re
from typing import Union

import numpy

TIMEDELTA_REGEX = (
    r"((?P<years>\d+)\s?(?:ys?|yrs?|years?))?\s*"
    r"((?P<months>\d+)\s?(?:mo|mons?|mths?|months?))?\s*"
    r"((?P<weeks>\d+)\s?(?:w|wks?|weeks?))?\s*"
    r"((?P<days>\d+)\s?(?:d|days?))?\s*"
    r"((?P<hours>\d+)\s?(?:h|hrs?|hours?))?\s*"
    r"((?P<minutes>\d+)\s?(?:m|mins?|minutes?))?\s*"
    r"((?P<seconds>\d+)\s?(?:s|secs?|seconds?))?\s*"
)

TIMEDELTA_PATTERN = re.compile(TIMEDELTA_REGEX, re.IGNORECASE)
UNIX_EPOCH: datetime.date = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)


def add_months(start_date: datetime.datetime, number_of_months: int):
    """
    Add months to a date, makes assumptions about how to handle the end of the month.
    """
    new_year, new_month = divmod(start_date.month - 1 + number_of_months, 12)
    new_year += start_date.year
    new_month += 1
    # Ensure the month is valid
    new_month = min(max(1, new_month), 12)
    last_day_of_month = (
        datetime.datetime(new_year, new_month % 12 + 1, 1) - datetime.timedelta(days=1)
    ).day
    new_day = min(start_date.day, last_day_of_month)
    return datetime.datetime(
        new_year,
        new_month,
        new_day,
        start_date.hour,
        start_date.minute,
        start_date.second,
        start_date.microsecond,
    )


def add_interval(
    current_date: datetime.datetime, interval: str
) -> Union[datetime.date, datetime.datetime]:
    """
    Parses a human readable timedelta (3d5h19m) into a datetime.timedelta.
    """
    match = TIMEDELTA_PATTERN.match(interval)
    if match:
        parts = {k: int(v) for k, v in match.groupdict().items() if v}
        # time delta doesn't include weeks, months or years
        if "weeks" in parts:
            weeks = parts.pop("weeks")
            current_date = current_date + datetime.timedelta(days=weeks * 7)
        if "months" in parts:
            months = parts.pop("months")
            current_date = add_months(current_date, months)
        if "years" in parts:
            # need to avoid 29th Feb problems, so can't just say year - year
            years = parts.pop("years")
            current_date = add_months(current_date, 12 * years)
        if parts:
            return current_date + datetime.timedelta(**parts)
        return current_date
    raise ValueError(f"Unable to interpret interval - {interval}")  # pragma: no cover


def date_range(start_date, end_date, interval: str):
    """Create a series of dates between two dates with a given interval"""
    start_date = parse_iso(start_date)
    end_date = parse_iso(end_date)

    if start_date > end_date:  # pragma: no cover
        raise ValueError("Cannot create an series with the provided start and end dates")

    # if the dates are the same, return that date
    if start_date == end_date:  # pragme: no cover
        yield start_date
        return

    cursor = start_date
    while cursor <= end_date:
        yield cursor
        cursor = add_interval(cursor, interval)


def parse_iso(value):
    # Date validation at speed is hard, dateutil is great but really slow, this is fast
    # but error-prone. It assumes it is a date or it really nothing like a date.
    # Making that assumption - and accepting the consequences - we can convert up to
    # three times faster than dateutil.
    #
    # valid formats (not exhaustive):
    #
    #   YYYY-MM-DD                 <- date
    #   YYYY-MM-DD HH:MM           <- date and time, no seconds
    #   YYYY-MM-DDTHH:MM           <- date and time, T separator
    #   YYYY-MM-DD HH:MM:SS        <- date and time with seconds
    #   YYYY-MM-DD HH:MM:SS.mmmm   <- date and time with milliseconds
    #
    # If the last character is a Z, we ignore it.
    # If we can't parse as a date we return None rather than error
    try:
        input_type = type(value)

        if input_type == numpy.datetime64:
            # this can create dates rather than datetimes, so don't return yet
            value = value.astype(datetime.datetime)
            input_type = type(value)
            if input_type is int:
                value /= 1000000000

        if input_type in (int, numpy.int64, float, numpy.float64):
            return datetime.datetime.fromtimestamp(int(value), tz=datetime.timezone.utc).replace(
                tzinfo=None
            )

        if input_type == datetime.datetime:
            return value.replace(microsecond=0)
        if input_type == datetime.date:
            return datetime.datetime.combine(value, datetime.time.min)

        # if we're here, we're doing string parsing
        if input_type == str and 10 <= len(value) <= 33:
            if value[-1] == "Z":
                value = value[:-1]
            if "+" in value:
                value = value.split("+")[0]
                if not 10 <= len(value) <= 28:
                    return None
            val_len = len(value)
            if value[4] != "-" or value[7] != "-":
                return None
            if val_len == 10:
                # YYYY-MM-DD
                return datetime.datetime(
                    *map(int, [value[:4], value[5:7], value[8:10]])  # type:ignore
                )
            if val_len >= 16:
                if value[10] not in ("T", " ") and value[13] != ":":
                    return None
                if val_len >= 19 and value[16] == ":":
                    # YYYY-MM-DD HH:MM:SS
                    return datetime.datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],  # YYYY
                                value[5:7],  # MM
                                value[8:10],  # DD
                                value[11:13],  # HH
                                value[14:16],  # MM
                                value[17:19],  # SS
                            ],
                        )
                    )
                if val_len == 16:
                    # YYYY-MM-DD HH:MM
                    return datetime.datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],
                                value[5:7],
                                value[8:10],
                                value[11:13],
                                value[14:16],
                            ],
                        )
                    )
        return None
    except (ValueError, TypeError) as r:
        return None


def date_trunc(truncate_to, date_value):
    """
    Truncate a datetime to a specified unit
    """
    date_value = parse_iso(date_value)

    if not isinstance(truncate_to, str):
        truncate_to = truncate_to[0]  # [#325]

    # [#711]
    truncate_to = str(truncate_to).lower()

    # fmt:off
    if truncate_to == "year":
        return datetime.datetime(date_value.year, 1, 1, tzinfo=date_value.tzinfo)
    elif truncate_to == "quarter":
        quarter = (date_value.month - 1) // 3 + 1
        return datetime.datetime(date_value.year, 3 * quarter - 2, 1, tzinfo=date_value.tzinfo)
    elif truncate_to == "month":
        return datetime.datetime(date_value.year, date_value.month, 1, tzinfo=date_value.tzinfo)
    elif truncate_to == "week":
        days_since_monday = date_value.weekday()
        monday = date_value - datetime.timedelta(days=days_since_monday)
        return date_trunc("day", monday)
    elif truncate_to == "day":
        return datetime.datetime(date_value.year, date_value.month, date_value.day, tzinfo=date_value.tzinfo)
    elif truncate_to == "hour":
        return datetime.datetime(date_value.year, date_value.month, date_value.day, date_value.hour, tzinfo=date_value.tzinfo)
    elif truncate_to == "minute":
        return datetime.datetime(date_value.year, date_value.month, date_value.day, date_value.hour, date_value.minute, tzinfo=date_value.tzinfo)
    elif truncate_to == "second":
        return datetime.datetime(date_value.year, date_value.month, date_value.day, date_value.hour, date_value.minute, date_value.second, tzinfo=date_value.tzinfo)
    else:
        raise ValueError("Invalid unit: {}".format(truncate_to))
    # fmt:on



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Ordering of adjacent predicates using a genetic algorithm.

This is only used for larger numbers of predicates - four or more.

For 5 predicates, we do 4 generations of 3 variations (12 rather than 25 variations tested)
For 10 predicates, we do 5 generations of 6 variations (30 rather than 100)
Fro 100 predicates, we do 8 generations of 51 variations (408 rather 10,000)

This may not find the _best_ solution, but it should find a better than average solution
"""
import random


def log2(x):
    if x <= 0:
        raise ValueError("Invalid input. Logarithm is undefined for non-positive numbers.")

    result = 0
    while x > 1:
        x /= 2
        result += 1

    return result


generations = log2
mutations = lambda x: (x // 2) + 1


def generate_initial_population(predicates):
    population = []
    population_size = (len(predicates) // 2) + 1

    for _ in range(population_size):
        arrangement = random.sample(predicates, len(predicates))
        population.append(arrangement)

    return population


def evaluate_cost(arrangement, cost_model):
    # Evaluate the cost of an arrangement using the cost model
    # You should implement this function based on your specific cost model
    # The lower the cost, the better the arrangement
    return cost_model(arrangement)


def mutate(arrangement):
    # Swap the order of two randomly selected predicates in the arrangement
    mutated_arrangement = arrangement[:]
    idx1, idx2 = random.sample(range(len(arrangement)), 2)
    mutated_arrangement[idx1], mutated_arrangement[idx2] = (
        mutated_arrangement[idx2],
        mutated_arrangement[idx1],
    )
    return mutated_arrangement


def genetic_algorithm(predicates, cost_model, population_size, num_generations, num_mutations):
    population = generate_initial_population(predicates)

    for _ in range(num_generations):
        # Evaluate the cost for each individual in the population
        costs = [evaluate_cost(individual, cost_model) for individual in population]

        # Sort the population by cost in ascending order
        population = [x for _, x in sorted(zip(costs, population))]
        fastest_so_far = population.pop(0)
        population = [fastest_so_far]

        # Mutate the current cheapest
        for _ in range(num_mutations):
            mutated_individual = mutate(fastest_so_far)
            population.append(mutated_individual)

    # Return the best arrangement found
    best_arrangement = population[0]
    best_cost = evaluate_cost(best_arrangement, cost_model)

    return best_arrangement, best_cost


# Example usage
predicates = [
    (100, 0.5),
    (5000, 0.1),
    (1000000, 0.01),
    (1000, 0.2),
    (10, 0.9),
]  # List of predicates (time, selectivity)
population_size = mutations(len(predicates))
num_generations = generations(len(predicates)) + 1
num_mutations = population_size


def cost_model(arrangement):
    # Define your cost model here
    # Calculate the cost of the arrangement based on your specific criteria
    print(arrangement)
    approx_records = 1000000
    cost = 0
    for time, selectivity in arrangement:
        cost += time * (time / approx_records)
        approx_records *= selectivity
    print(cost, approx_records)
    return cost


best_arrangement, best_cost = genetic_algorithm(
    predicates, cost_model, population_size, num_generations, num_mutations
)

print("Best Arrangement:", best_arrangement)
print("Best Cost:", best_cost)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Global Buffer Pool.

This is little more than a wrapper around the LRU-K(2) cache.
"""
from array import array
from typing import Any
from typing import Optional

from opteryx.utils.lru_2 import LRU2


class NullCacheBackEnd:
    """
    We can remove a check in each operation by just having a null service.
    """

    def get(self, key: bytes) -> None:
        return None

    def set(self, key: bytes, value: Any) -> None:
        return None


class _BufferPool:
    """
    Buffer Pool is a class implementing a Least Recently Used (LRU) cache of buffers.
    """

    slots = "_lru", "_cache_backend", "_max_cacheable_item_size"

    def __init__(self, cache_manager):
        self._cache_backend = cache_manager.cache_backend
        if not self._cache_backend:
            self._cache_backend = (
                NullCacheBackEnd()
            )  # rather than make decisions - just use a dummy
        self.max_cacheable_item_size = cache_manager.max_cacheable_item_size
        self._lru = LRU2(size=cache_manager.max_local_buffer_capacity)

    def get(self, key: bytes) -> Optional[array]:
        """
        Retrieve an item from the pool, return None if the item isn't found.
        If cache is provided and item is not in pool, attempt to get it from cache.
        """
        value = self._lru.get(key)
        if value is not None:
            return value
        return self._cache_backend.get(key)

    def set(self, key: bytes, value) -> Optional[str]:
        """
        Put an item into the pool, evict an item if the pool is full.
        If a cache is provided, also set the value in the cache.
        """
        evicted = self._lru.set(key, value)
        self._cache_backend.set(key, value)
        return evicted

    @property
    def stats(self) -> tuple:
        """
        Return the hit, miss and eviction statistics for the buffer pool.
        """
        return self._lru.stats

    def reset(self, reset_stats: bool = False):
        """
        Reset the buffer pool.
        If reset_stats is True, also reset the statistics.
        """
        self._lru.reset(reset_stats=reset_stats)


class BufferPool(_BufferPool):
    """
    Singleton wrapper for the _BufferPool class. Only allows one instance of _BufferPool to exist.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = cls._create_instance()
        return cls._instance

    @classmethod
    def _create_instance(cls):
        # Import here to avoid circular imports
        from opteryx import get_cache_manager

        cache_manager = get_cache_manager()

        return _BufferPool(cache_manager)

    @classmethod
    def reset(cls):
        """
        Reset the BufferPool singleton instance. This is useful when the configuration changes.
        """
        cls._instance = None
        cls._instance = cls._create_instance()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is a temporary step, which takes logical plans from the V2 planner
and converts them to modified-V1 physical plans.

This should look different when the operators are rewritten for the 
Gen 2 execution engine (a later piece of work)
"""

from opteryx import operators
from opteryx.components.logical_planner import LogicalPlanStepType
from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.models import ExecutionTree


def create_physical_plan(logical_plan, query_properties):
    plan = ExecutionTree()

    for nid, logical_node in logical_plan.nodes(data=True):
        node_type = logical_node.node_type
        node_config = logical_node.properties
        node: operators.BasePlanNode = None

        # fmt: off
        if node_type == LogicalPlanStepType.Aggregate:
            node = operators.AggregateNode(query_properties, aggregates=node_config["aggregates"])
        elif node_type == LogicalPlanStepType.AggregateAndGroup:
            node = operators.AggregateAndGroupNode(query_properties, groups=node_config["groups"], aggregates=node_config["aggregates"], projection=node_config["projection"])
        elif node_type == LogicalPlanStepType.Defragment:
            node = operators.MorselDefragmentNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Distinct:
            node = operators.DistinctNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Exit:
            node = operators.ExitNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Explain:
            node = operators.ExplainNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Filter:
            node = operators.SelectionNode(query_properties, filter=node_config["condition"])
        elif node_type == LogicalPlanStepType.FunctionDataset:
            node = operators.FunctionDatasetNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Join:
            if node_config.get("type") == "cross join":
                # CROSS JOINs are quite different
                node = operators.CrossJoinNode(query_properties, **node_config)
            else:
                node = operators.JoinNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Limit:
            node = operators.LimitNode(query_properties, limit=node_config.get("limit"), offset=node_config.get("offset", 0))
        elif node_type == LogicalPlanStepType.Order:
            node = operators.SortNode(query_properties, order=node_config["order_by"])
        elif node_type == LogicalPlanStepType.Project:
            node = operators.ProjectionNode(query_properties, projection=logical_node.columns)
        elif node_type == LogicalPlanStepType.Scan:
            node = operators.ScannerNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Set:
            node = operators.SetVariableNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Show:
            if node_config["items"][0] == "PARAMETER":
                node = operators.ShowValueNode(query_properties, kind="PARAMETER", value=node_config["items"][1])
            else:
                raise UnsupportedSyntaxError("Can only SHOW variables")
        elif node_type == LogicalPlanStepType.ShowColumns:
            node = operators.ShowColumnsNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Subquery:
            node = operators.NoOpNode(query_properties, **node_config)
        elif node_type == LogicalPlanStepType.Union:
            node = operators.UnionNode(query_properties, **node_config)

        else:
            raise Exception(f"something unexpected happed - {node_type.name}")
        # fmt: on
        plan.add_node(nid, node)

    for source, destination, relation in logical_plan.edges():
        plan.add_edge(source, destination)

    return plan



import numpy as np

from cjoin import cython_inner_join
from cjoin import cython_left_join

from .helpers import columns_to_array
from .helpers import groupify_array


def align_tables(source_table, append_table, source_indices, append_indices):
    # If either source_indices or append_indices is empty, return the source_table taken with source_indices immediately
    if len(source_indices) == 0 or len(append_indices) == 0:
        return source_table.take(source_indices)

    # Take the rows from source_table at the specified source_indices
    aligned_table = source_table.take(source_indices)

    # Create a set of column names from the source table for efficient existence checking
    source_column_names = set(source_table.column_names)

    # Iterate through the column names of append_table
    for column_name in append_table.column_names:
        # If the column_name is not found in source_column_names
        if column_name not in source_column_names:
            # Append the column from append_table to aligned_table, taking the elements at the specified append_indices
            aligned_table = aligned_table.append_column(
                column_name, append_table.column(column_name).take(append_indices)
            )

    # Return the aligned table
    return aligned_table


def preprocess_left_side(left, left_on):
    l_array = columns_to_array(left, left_on)
    l_distinct, lc, l_sort_idxs, lbi = groupify_array(l_array)
    return l_distinct, lc, l_sort_idxs.astype(np.int64), lbi


def inner_join_with_preprocessed_left(left, right, right_on, l_distinct, lc, l_sort_idxs, lbi):
    # Gather join columns - create arrays of the hashes of the values in the column
    # updated for Opteryx
    r_array = columns_to_array(right, right_on)

    # Groupify the join array, this generates a set of data about the array
    # including the unique values in the array, and the sort order for the array.
    r_distinct, rc, r_sort_idxs, rbi = groupify_array(r_array)

    # Create the list of unique values combining the column from the left and the right
    # tables
    unique, inv = np.unique(np.concatenate([l_distinct, r_distinct]), return_inverse=True)

    # Align Left side
    # the inv array the positions in the unique list of the combined left and right list,
    # because we build this using np.concat, we know the first set of records is from the
    # left list.
    linv = inv[: l_distinct.shape[0]]
    # this creates empty masks
    lcc, lbic = np.zeros_like(unique), np.zeros_like(unique)
    # this sets the values at the positions in linv to the count values from groupify above
    lcc[linv] = lc
    # this sets the values at the positions in linv to the begin indexes of the groups
    lbic[linv] = lbi

    # Align right side
    # the inv array the positions in the unique list of the combined left and right list,
    # because we build this using np.concat, we know the end set of records is from the
    # right list.
    rinv = inv[l_distinct.shape[0] :]
    # this creates empty masks
    rcc, rbic = np.zeros_like(unique), np.zeros_like(unique)
    # this sets the values at the positions in rinv to the count values from groupify above
    rcc[rinv] = rc
    # this sets the values at the positions in rinv to the begin indexes of the groups
    rbic[rinv] = rbi

    # Perform cjoin
    left_align, right_align = cython_inner_join(
        l_sort_idxs,
        r_sort_idxs,
        lcc,
        rcc,
        lbic,
        rbic,
    )

    return align_tables(left, right, left_align, right_align)


def inner_join(left, right, left_on, right_on):
    # Gather join columns - create arrays of the hashes of the values in the column
    # updated for Opteryx
    l_array, r_array = columns_to_array(left, left_on), columns_to_array(right, right_on)

    # Groupify the join array, this generates a set of data about the array
    # including the unique values in the array, and the sort order for the array.
    l_distinct, lc, l_sort_idxs, lbi = groupify_array(l_array)
    r_distinct, rc, r_sort_idxs, rbi = groupify_array(r_array)

    # Create the list of unique values combining the column from the left and the right
    # tables
    unique, inv = np.unique(np.concatenate([l_distinct, r_distinct]), return_inverse=True)

    # Align Left side
    # the inv array the positions in the unique list of the combined left and right list,
    # because we build this using np.concat, we know the first set of records is from the
    # left list.
    linv = inv[: l_distinct.shape[0]]
    # this creates empty masks
    lcc, lbic = np.zeros_like(unique), np.zeros_like(unique)
    # this sets the values at the positions in linv to the count values from groupify above
    lcc[linv] = lc
    # this sets the values at the positions in linv to the begin indexes of the groups
    lbic[linv] = lbi

    # Align right side
    # the inv array the positions in the unique list of the combined left and right list,
    # because we build this using np.concat, we know the end set of records is from the
    # right list.
    rinv = inv[l_distinct.shape[0] :]
    # this creates empty masks
    rcc, rbic = np.zeros_like(unique), np.zeros_like(unique)
    # this sets the values at the positions in rinv to the count values from groupify above
    rcc[rinv] = rc
    # this sets the values at the positions in rinv to the begin indexes of the groups
    rbic[rinv] = rbi

    # Perform cjoin
    left_align, right_align = cython_inner_join(
        l_sort_idxs,
        r_sort_idxs,
        lcc,
        rcc,
        lbic,
        rbic,
    )

    return align_tables(left, right, left_align, right_align)


def left_join(left, right, left_on, right_on):  # pragma: no cover - currently not called
    # Gather join columns - create arrays of the hashes of the values in the column
    # new for Opteryx
    l_array, r_array = columns_to_array(left, left_on), columns_to_array(right, right_on)

    # Groupify the join array, this generates a set of data about the array
    # including the unique values in the array, and the sort order for the array.
    l_distinct, lc, l_sort_idxs, lbi = groupify_array(l_array)
    r_distinct, rc, r_sort_idxs, rbi = groupify_array(r_array)

    # Create the list of unique values combining the column from the left and the right
    # tables
    unique, inv = np.unique(np.concatenate([l_distinct, r_distinct]), return_inverse=True)

    # Align Left side
    # the inv array the positions in the unique list of the combined left and right list,
    # because we build this using np.concat, we know the first set of records is from the
    # left list.
    linv = inv[: l_distinct.shape[0]]
    # this creates empty masks
    lcc, lbic = np.zeros_like(unique), np.zeros_like(unique)
    # this sets the values at the positions in linv to the count values from groupify above
    lcc[linv] = lc
    # this sets the values at the positions in linv to the begin indexes of the groups
    lbic[linv] = lbi

    # Align right side
    # the inv array the positions in the unique list of the combined left and right list,
    # because we build this using np.concat, we know the end set of records is from the
    # right list.
    rinv = inv[l_distinct.shape[0] :]
    # this creates empty masks
    rcc, rbic = np.zeros_like(unique), np.zeros_like(unique)
    # this sets the values at the positions in rinv to the count values from groupify above
    rcc[rinv] = rc
    # this sets the values at the positions in rinv to the begin indexes of the groups
    rbic[rinv] = rbi

    rows = len(l_array) * len(r_array)
    left_align, right_align = np.empty(rows, dtype=np.int64), np.empty(rows, dtype=np.int64)

    # Perform cjoin
    left_align, right_align = cython_left_join(
        l_sort_idxs.astype(np.int64),
        r_sort_idxs.astype(np.int64),
        lcc.astype(np.int64),
        rcc.astype(np.int64),
        lbic.astype(np.int64),
        rbic.astype(np.int64),
    )

    return align_tables(left, right, left_align, right_align)



import os
import threading
from time import sleep

import orjson

RESOURCE_LIB = True
try:
    import resource
except ImportError:  # pragma: no cover
    RESOURCE_LIB = False
    import psutil  # type:ignore


class ResourceMonitor:  # pragma: no cover
    slots = "frequency"

    def __init__(self, frequency=0.01):  # pragma: no cover
        self.frequency = frequency
        if not RESOURCE_LIB:
            self.process = psutil.Process(os.getpid())
        else:
            print(resource.getrlimit(resource.RUSAGE_SELF))

    def resource_usage(self):  # pragma: no cover
        while True:
            if RESOURCE_LIB:
                memory_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
            else:
                memory_usage = self.process.memory_info()[0]
            print(orjson.dumps({"memory": memory_usage}), RESOURCE_LIB)
            sleep(self.frequency)


monitor = ResourceMonitor()
resource_thread = threading.Thread(target=monitor.resource_usage)
resource_thread.name = "mabel-resource-monitor"
resource_thread.daemon = True
resource_thread.start()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is Binder, it sits between the Logical Planner and the Optimizers.

~~~
                      ┌───────────┐
                      │   USER    │
         ┌────────────┤           ◄────────────┐
         │SQL         └───────────┘            │
  ───────┼─────────────────────────────────────┼──────
         │                                     │
   ┌─────▼─────┐                               │
   │ SQL       │                               │
   │  Rewriter │                               │
   └─────┬─────┘                               │
         │SQL                                  │Results
   ┌─────▼─────┐                         ┌─────┴─────┐
   │           │                         │           │
   │ Parser    │                         │ Executor  │
   └─────┬─────┘                         └─────▲─────┘
         │AST                                  │Plan
   ┌─────▼─────┐                         ╔═══════════╗
   │ AST       │                         ║Cost-Based ║
   │ Rewriter  │                         ║ Optimizer ║
   └─────┬─────┘                         ╚═════▲═════╝
         │AST                                  │Plan
   ┌─────▼─────┐      ┌───────────┐      ┌─────┴─────┐
   │ Logical   │ Plan │ Heuristic │ Plan │           │
   │   Planner ├──────► Optimizer ├──────► Binder    │
   └───────────┘      └───────────┘      └─────▲─────┘
                                               │Schemas
                                         ┌─────┴─────┐
                                         │           │
                                         │ Catalogue │
                                         └───────────┘
~~~

The Binder is responsible for adding information about the database and engine into the
Logical Plan.

The binder takes the the logical plan, and adds information from various catalogues
into that planand then performs some validation checks.

These catalogues include:
- The Data Catalogue (e.g. data schemas)
- The Function Catalogue (e.g. function inputs and data types)
- The Variable Catalogue (i.e. the @ variables)

The Binder performs these activities:
- schema lookup and propagation (add columns and types, add aliases)

"""


from opteryx.components.binder.binder_visitor import BinderVisitor
from opteryx.components.binder.binding_context import BindingContext
from opteryx.components.logical_planner import LogicalPlan
from opteryx.exceptions import InvalidInternalStateError


def do_bind_phase(plan: LogicalPlan, connection=None, qid: str = None) -> LogicalPlan:
    """
    Execute the bind phase of the query engine.

    Parameters:
        plan: Any
            The logical plan.
        context: BindingContext
            The context needed for the binding phase.

    Returns:
        Modified logical plan after the binding phase.

    Raises:
        InvalidInternalStateError: Raised when the logical plan has more than one root node.
    """
    binder_visitor = BinderVisitor()
    root_node = plan.get_exit_points()
    context = BindingContext.initialize(qid=qid, connection=connection)

    if len(root_node) > 1:
        raise InvalidInternalStateError(
            f"{context.qid} - logical plan has {len(root_node)} heads - this is an error"
        )

    plan, _ = binder_visitor.traverse(plan, root_node[0], context=context)

    # DEBUG: log ("AFTER BINDING")
    # DEBUG: log (plan.draw())

    return plan



import functools
import textwrap
from collections import defaultdict
from typing import NamedTuple


class Query:
    keywords = [
        "WITH",
        "SELECT",
        "FROM",
        "WHERE",
        "GROUP BY",
        "HAVING",
        "ORDER BY",
        "LIMIT",
    ]

    separators = dict(WHERE="AND", HAVING="AND")
    default_separator = ","

    formats: tuple = (
        defaultdict(lambda: "{value}"),
        defaultdict(lambda: "{value} AS {alias}", WITH="{alias} AS {value}"),
    )

    subquery_keywords = {"WITH"}
    fake_keywords = dict(JOIN="FROM")
    flag_keywords = dict(SELECT={"DISTINCT", "ALL"})

    def __init__(self, data=None, separators=None):
        self.data = {}
        if data is None:
            data = dict.fromkeys(self.keywords, ())
        for keyword, args in data.items():
            self.data[keyword] = _FlagList()
            self.add(keyword, *args)

        if separators is not None:
            self.separators = separators

    def add(self, keyword, *args):
        keyword, fake_keyword = self._resolve_fakes(keyword)
        keyword, flag = self._resolve_flags(keyword)
        target = self.data[keyword]

        if flag:
            if target.flag:
                raise ValueError(f"{keyword} already has flag: {flag!r}")
            target.flag = flag

        kwargs: dict = {}
        if fake_keyword:
            kwargs.update(keyword=fake_keyword)
        if keyword in self.subquery_keywords:
            kwargs.update(is_subquery=True)

        for arg in args:
            target.append(_Thing.from_arg(arg, **kwargs))

        return self

    def _resolve_fakes(self, keyword):
        for part, real in self.fake_keywords.items():
            if part in keyword:
                return real, keyword
        return keyword, ""

    def _resolve_flags(self, keyword):
        prefix, _, flag = keyword.partition(" ")
        if prefix in self.flag_keywords:
            if flag and flag not in self.flag_keywords[prefix]:
                raise ValueError(f"invalid flag for {prefix}: {flag!r}")
            return prefix, flag
        return keyword, ""

    def __getattr__(self, name):
        if not name.isupper():
            return getattr(super(), name)
        return functools.partial(self.add, name.replace("_", " "))

    def __str__(self):
        return "".join(self._lines())

    def _lines(self):
        for keyword, things in self.data.items():
            if not things:
                continue

            if things.flag:
                yield f"{keyword} {things.flag}\n"
            else:
                yield f"{keyword}\n"

            grouped: tuple = [], []
            for thing in things:
                grouped[bool(thing.keyword)].append(thing)
            for group in grouped:
                yield from self._lines_keyword(keyword, group)

    def _lines_keyword(self, keyword, things):
        for i, thing in enumerate(things, 1):
            last = i == len(things)

            if thing.keyword:
                yield thing.keyword + "\n"

            format_ = self.formats[bool(thing.alias)][keyword]
            value = thing.value
            if thing.is_subquery:
                value = f"(\n{self._indent(value)}\n)"
            yield self._indent(format_.format(value=value, alias=thing.alias))

            if not last and not thing.keyword:
                try:
                    yield " " + self.separators[keyword]
                except KeyError:
                    yield self.default_separator

            yield "\n"

    _indent = functools.partial(textwrap.indent, prefix="    ")


class _Thing(NamedTuple):
    value: str
    alias: str = ""
    keyword: str = ""
    is_subquery: bool = False

    @classmethod
    def from_arg(cls, arg, **kwargs):
        if isinstance(arg, str):
            alias, value = "", arg
        elif len(arg) == 2:
            alias, value = arg
        else:
            raise ValueError(f"invalid arg: {arg!r}")
        return cls(_clean_up(value), _clean_up(alias), **kwargs)


class _FlagList(list):
    flag: str = ""


def _clean_up(thing: str) -> str:
    return textwrap.dedent(thing.rstrip()).strip()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Arrow Reader

Used to read datasets registered using the register_arrow or register_df functions.
"""

import pyarrow
from orso.schema import FlatColumn
from orso.schema import RelationSchema

from opteryx.connectors.base.base_connector import DEFAULT_MORSEL_SIZE
from opteryx.connectors.base.base_connector import BaseConnector
from opteryx.shared import MaterializedDatasets
from opteryx.utils import arrow


class ArrowConnector(BaseConnector):
    __mode__ = "Internal"

    def __init__(self, *args, **kwargs):
        BaseConnector.__init__(self, **kwargs)

        self.dataset = self.dataset.lower()
        self._datasets = MaterializedDatasets()

    def get_dataset_schema(self) -> RelationSchema:
        dataset = self._datasets[self.dataset]
        arrow_schema = dataset.schema

        self.schema = RelationSchema(
            name=self.dataset,
            columns=[FlatColumn.from_arrow(field) for field in arrow_schema],
        )

        return self.schema

    def read_dataset(self, columns: list = None, **kwargs) -> pyarrow.Table:
        dataset = self._datasets[self.dataset]

        batch_size = DEFAULT_MORSEL_SIZE // (dataset.nbytes / dataset.num_rows)

        for batch in dataset.to_batches(max_chunksize=batch_size):
            morsel = pyarrow.Table.from_batches([batch], schema=dataset.schema)
            if columns:
                morsel = arrow.post_read_projector(morsel, columns)
            yield morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Show Variables Node

This is a SQL Query Execution Plan Node.
"""
from typing import Generator

from opteryx.constants import QueryStatus
from opteryx.models import NonTabularResult
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


class SetVariableNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)

        self.variable = config.get("variable")
        self.value = config.get("value")

        self.variables = config.get("variables")

    @property
    def name(self):  # pragma: no cover
        return "Set Variables"

    @property
    def config(self):  # pragma: no cover
        return f"{self.variable} TO {self.value}"

    def execute(self) -> Generator:
        self.variables[self.variable] = self.value
        return NonTabularResult(record_count=1, status=QueryStatus.SQL_SUCCESS)  # type: ignore



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import urllib.request
from typing import Dict
from typing import List

import pyarrow
import requests
from google.auth.transport.requests import Request
from orso.schema import RelationSchema
from orso.tools import single_item_cache
from orso.types import OrsoTypes

from opteryx.connectors.base.base_connector import BaseConnector
from opteryx.connectors.capabilities import Cacheable
from opteryx.connectors.capabilities import Partitionable
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.exceptions import DatasetNotFoundError
from opteryx.exceptions import MissingDependencyError
from opteryx.exceptions import UnsupportedFileTypeError
from opteryx.utils import paths
from opteryx.utils.file_decoders import TUPLE_OF_VALID_EXTENSIONS
from opteryx.utils.file_decoders import get_decoder

OS_SEP = os.sep


class GcpCloudStorageConnector(BaseConnector, Cacheable, Partitionable, PredicatePushable):
    __mode__ = "Blob"

    PUSHABLE_OPS: Dict[str, bool] = {
        "Eq": True,
        "NotEq": True,
        "Gt": True,
        "GtEq": True,
        "Lt": True,
        "LtEq": True,
    }

    PUSHABLE_TYPES = {OrsoTypes.BOOLEAN, OrsoTypes.DOUBLE, OrsoTypes.INTEGER, OrsoTypes.VARCHAR}

    def __init__(self, credentials=None, **kwargs):
        try:
            from google.auth.credentials import AnonymousCredentials
            from google.cloud import storage
        except ImportError as err:
            raise MissingDependencyError(err.name) from err

        BaseConnector.__init__(self, **kwargs)
        Partitionable.__init__(self, **kwargs)
        Cacheable.__init__(self, **kwargs)
        PredicatePushable.__init__(self, **kwargs)

        self.dataset = self.dataset.replace(".", OS_SEP)
        self.credentials = credentials
        self.bucket, _, _, _ = paths.get_parts(self.dataset)

        # we're going to cache the first blob as the schema and dataset reader
        # sometimes both start here
        self.client = self._get_storage_client()
        self.client_credentials = self.client._credentials

        # Cache access tokens for accessing GCS
        if not self.client_credentials.valid:
            request = Request()
            self.client_credentials.refresh(request)
            self.access_token = self.client_credentials.token

        # Create a HTTP connection session to reduce effort for
        # each fetch
        self.session = requests.Session()

    def _get_storage_client(self):
        from google.cloud import storage

        if os.environ.get("STORAGE_EMULATOR_HOST"):
            from google.auth.credentials import AnonymousCredentials

            return storage.Client(credentials=AnonymousCredentials())
        else:  # pragma: no cover
            return storage.Client()

    def read_blob(self, *, blob_name, **kwargs):
        # For performance we use the GCS API directly, this is roughly 10%
        # faster than using the SDK. As one of the slowest parts of the system
        # 10% can be measured in seconds.
        bucket, _, _, _ = paths.get_parts(blob_name)

        # Ensure the credentials are valid, refreshing them if necessary
        if not self.client_credentials.valid:
            request = Request()
            self.client_credentials.refresh(request)
            self.access_token = self.client_credentials.token

        bucket = bucket.replace("va_data", "va-data")
        bucket = bucket.replace("data_", "data-")
        object_full_path = urllib.parse.quote(blob_name[(len(bucket) + 1) :], safe="")

        url = f"https://storage.googleapis.com/storage/v1/b/{bucket}/o/{object_full_path}?alt=media"

        response = self.session.get(
            url, headers={"Authorization": f"Bearer {self.access_token}"}, timeout=30
        )
        return response.content

    @single_item_cache
    def get_list_of_blob_names(self, *, prefix: str) -> List[str]:
        bucket, object_path, _, _ = paths.get_parts(prefix)
        bucket = bucket.replace("va_data", "va-data")
        bucket = bucket.replace("data_", "data-")

        object_path = urllib.parse.quote(object_path, safe="")
        bucket = urllib.parse.quote(bucket, safe="")  # Ensure bucket name is URL-safe
        url = f"https://storage.googleapis.com/storage/v1/b/{bucket}/o?prefix={object_path}&fields=items(name)"

        # Ensure the credentials are valid, refreshing them if necessary
        if not self.client_credentials.valid:
            request = Request()
            self.client_credentials.refresh(request)
            self.access_token = self.client_credentials.token

        headers = {"Authorization": f"Bearer {self.access_token}"}
        response = self.session.get(url, headers=headers, timeout=30)

        if response.status_code != 200:
            raise Exception(f"Error fetching blob list: {response.text}")

        blob_data = response.json()
        blob_names = sorted(
            f"{bucket}/{name}"
            for name in (blob["name"] for blob in blob_data.get("items", []))
            if name.endswith(TUPLE_OF_VALID_EXTENSIONS)
        )

        return blob_names

    def read_dataset(
        self, columns: list = None, predicates: list = None, just_schema: bool = False, **kwargs
    ) -> pyarrow.Table:
        blob_names = self.partition_scheme.get_blobs_in_partition(
            start_date=self.start_date,
            end_date=self.end_date,
            blob_list_getter=self.get_list_of_blob_names,
            prefix=self.dataset,
        )

        for blob_name in blob_names:
            try:
                decoder = get_decoder(blob_name)
                blob_bytes = self.read_blob(blob_name=blob_name, statistics=self.statistics)
                yield decoder(
                    blob_bytes, projection=columns, selection=predicates, just_schema=just_schema
                )
            except UnsupportedFileTypeError:
                pass

    def get_dataset_schema(self) -> RelationSchema:
        # Try to read the schema from the metastore
        self.schema = self.read_schema_from_metastore()
        if self.schema:
            return self.schema

        # Read first blob for schema inference and cache it
        self.schema = next(self.read_dataset(just_schema=True), None)

        if self.schema is None:
            raise DatasetNotFoundError(dataset=self.dataset)

        return self.schema



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Opteryx and Orso Import Customization
-------------------------------------
This module provides custom import mechanics for modules within Opteryx and Orso projects.
In production, Python's import system will see the original source code without comments.
In a development environment, specific comments like '# DEBUG: log' are transformed into
print statements, providing additional debugging information.

The custom import system utilizes Python's importlib and AST manipulation to achieve this.
"""

# Import only necessary for class inheritance, minimizes startup overhead
import datetime
import importlib.abc
import importlib.util
import sys


class OpteryxOrsoImportFinder(importlib.abc.MetaPathFinder):
    """
    Custom Import Finder that looks for modules starting with 'opteryx' or 'Orso'.
    """

    def find_spec(self, fullname, path, target=None):
        if not (fullname.startswith("opteryx") or fullname.startswith("Orso")):
            return None

        sys.meta_path.remove(self)
        spec = importlib.util.find_spec(fullname)
        sys.meta_path.insert(0, self)

        if spec is None:
            return None

        if spec.loader and isinstance(spec.loader, importlib.abc.SourceLoader):
            spec.loader = OpteryxOrsoImportLoader(spec.loader)

        return spec


class OpteryxOrsoImportLoader(importlib.abc.SourceLoader):
    """
    Custom Import Loader to process Python source code before it's actually imported.
    """

    def __init__(self, original_loader):
        self.original_loader = original_loader

    def get_filename(self, fullname):
        return self.original_loader.get_filename(fullname)

    def get_data(self, filepath: str) -> bytes:
        # Delayed import to minimize startup overhead
        from pathlib import Path

        try:
            file_extension = Path(filepath).suffix

            if file_extension != ".py":
                return self.original_loader.get_data(filepath)

            with open(filepath, "r", encoding="utf-8") as f:
                original_source = f.read()

            cleaned_source = self._remove_comments(original_source)
            return cleaned_source.encode("utf-8")

        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    def _remove_comments(self, source: str) -> str:
        debug_mode_enabled = source.replace("# DEBUG: log", "print")
        debug_mode_enabled = debug_mode_enabled.replace("# DEBUG: ", "")
        return debug_mode_enabled


# Register the custom MetaPathFinder
print(f"{datetime.datetime.now()} [LOADER] Loading Opteryx in DEBUG mode.")
sys.meta_path.insert(0, OpteryxOrsoImportFinder())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy
from orso.types import OrsoTypes

from opteryx.exceptions import InvalidFunctionParameterError
from opteryx.exceptions import SqlError
from opteryx.utils import dates


def generate_series(*args):
    arg_len = len(args)
    arg_vals = [i.value for i in args]
    first_arg_type = args[0].type

    # if the parameters are numbers, generate series is an alias for range
    if first_arg_type in (
        OrsoTypes.INTEGER,
        OrsoTypes.DOUBLE,
    ):
        if arg_len not in (1, 2, 3):
            raise SqlError(
                "generate_series for numbers takes 1 (stop), 2 (start, stop) or 3 (start, stop, interval) parameters."
            )
        return numeric_range(*arg_vals)

    # if the params are timestamps, we create time intervals
    if args[0].type in (OrsoTypes.DATE, OrsoTypes.TIMESTAMP):
        if arg_len != 3:
            raise SqlError("generate_series for dates needs start, end, and interval parameters")
        return dates.date_range(*arg_vals)

    raise InvalidFunctionParameterError(
        "Unsupported value for GENERATE_SERIES, must be date or numeric series."
    )


def numeric_range(*args) -> numpy.ndarray:
    """
    Generate a numeric range of vales

    Args:
        [start, ]stop, [step, ]: Arguments as in numpy.arange.

    Returns:
        numpy.ndarray: Array of evenly spaced values.

    Raises:
        ValueError: If the number of arguments is not 1, 2, or 3.

    Examples:
        generate_range(5)
        generate_range(1, 5)
        generate_range(1, 5, 0.5)
    """
    # Define defaults
    start, step, dtype = numpy.int64(1), numpy.int64(1), numpy.float64

    # Process arguments
    if len(args) == 1:
        stop = args[0]
    elif len(args) == 2:
        start, stop = args
    elif len(args) == 3:
        start, stop, step = args
    else:
        raise ValueError("Invalid number of arguments. Expected 1, 2, or 3: start, stop [, step].")

    # Determine dtype
    if all(numpy.issubdtype(arg, numpy.integer) for arg in [start, stop, step]):
        dtype = numpy.int64  # type: ignore

    # Compute range
    num_range = numpy.arange(start, stop + step, step, dtype=dtype)

    # Check last value, remove if it doesn't fall on a step boundary or is over the stop value
    if not numpy.isclose(num_range[-1], stop, atol=step / 2) or num_range[-1] > stop:  # type: ignore
        num_range = num_range[:-1]

    return num_range



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orso.tools import random_string
from orso.types import PYTHON_TO_ORSO_MAP

from opteryx.components.logical_planner import LogicalPlan
from opteryx.components.logical_planner import LogicalPlanNode
from opteryx.components.logical_planner import LogicalPlanStepType
from opteryx.managers.expression import ORSO_TO_NUMPY_MAP
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import get_all_nodes_of_type

from .optimization_strategy import HeuristicOptimizerContext
from .optimization_strategy import OptimizationStrategy

REWRITES = {"InList": "Eq", "NotInList": "NotEq"}


def _rewriter(node):

    if node.value in REWRITES:
        if node.right.node_type == NodeType.LITERAL and len(node.right.value) == 1:
            node.value = REWRITES[node.value]
            node.right.value = node.right.value.pop()
            node.right.type = node.right.sub_type
            node.right.sub_type = None

    elif node.node_type in (NodeType.AND, NodeType.OR):
        node.left = _rewriter(node.left)
        node.right = _rewriter(node.right)

    return node


class RewriteInWithSingleComparitorStrategy(OptimizationStrategy):
    def visit(
        self, node: LogicalPlanNode, context: HeuristicOptimizerContext
    ) -> HeuristicOptimizerContext:
        """
        Rewrite IN conditions with a single comparitor, e.g. a IN (1), to
        an Equals condition (or Not Equals). Equals conditions can be
        pushed into more places than IN conditions.
        """

        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if node.node_type == LogicalPlanStepType.Filter:
            node.condition = _rewriter(node.condition)
            context.optimized_plan[context.node_id] = node

        return context

    def complete(self, plan: LogicalPlan, context: HeuristicOptimizerContext) -> LogicalPlan:
        # No finalization needed for this strategy
        return plan



from .constant_folding import ConstantFoldingStrategy
from .predicate_pushdown import PredicatePushdownStrategy
from .projection_pushdown import ProjectionPushdownStrategy

__all__ = [
    "ConstantFoldingStrategy",
    "PredicatePushdownStrategy",
    "ProjectionPushdownStrategy",
]


def _tag_predicates(nodes):
    from opteryx.managers.expression import NodeType

    NODE_ORDER: dict = {}
    """
    Here we add tags to the predicates to assist with optimization.

    Weighting of predicates based on naive rules, this is mostly useful for situations where
    we do not have statistics to make cost-based decisions later. We're going to start with
    arbitrary numbers, we need to find a way to refine these over time. The logic is
    roughly:
        - 35 is something that is expensive (we're running function)
        - 32 is where we're doing a complex comparison
        ...
        - 7 are IS/IS NOT filters
        - 5 is doing an eqi comparison on a column and a literal
        - 3 is doing an eqi comparison on two literals (don't actually do that in a filter)
    """

    for node in nodes:
        while node.node_type == NodeType.NESTED and node.centre:
            node = node.centre

        node.weight = 0
        node.simple = True
        node.relations = set()

        if node.condition.node_type == NodeType.UNARY_OPERATOR:
            # these are IS/IS NOT filters
            node.weight += 7
            node.relations.add(node.condition.centre.source)
            continue
        if not node.condition.node_type == NodeType.COMPARISON_OPERATOR:
            node.weight += 35
            node.simple = False
            continue
        node.weight = NODE_ORDER.get(node.condition.value, 12)
        if node.condition.left.node_type == NodeType.LITERAL:
            node.weight += 1
        elif node.condition.left.node_type == NodeType.IDENTIFIER:
            node.weight += 3
            node.relations.add(node.condition.left.source)
        else:
            node.weight += 10
            node.simple = False
        if node.condition.right.node_type == NodeType.LITERAL:
            node.weight += 1
        elif node.condition.right.node_type == NodeType.IDENTIFIER:
            node.weight += 3
            node.relations.add(node.condition.right.source)
        else:
            node.weight += 10
            node.simple = False

    return sorted(nodes, key=lambda node: node.weight, reverse=True)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The file connector provides the reader for when a file name is provided as the
dataset name in a query.
"""
from typing import Dict
from typing import Optional

import pyarrow
from orso.schema import RelationSchema
from orso.types import OrsoTypes

from opteryx.connectors.base.base_connector import BaseConnector
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.exceptions import DatasetNotFoundError
from opteryx.utils.file_decoders import get_decoder


class FileConnector(BaseConnector, PredicatePushable):
    __mode__ = "Blob"
    _byte_array: Optional[bytes] = None  # Instance attribute to store file bytes

    PUSHABLE_OPS: Dict[str, bool] = {
        "Eq": True,
        "NotEq": True,
        "Gt": True,
        "GtEq": True,
        "Lt": True,
        "LtEq": True,
    }

    PUSHABLE_TYPES = {OrsoTypes.BOOLEAN, OrsoTypes.DOUBLE, OrsoTypes.INTEGER, OrsoTypes.VARCHAR}

    @property
    def interal_only(self):
        return True

    def __init__(self, *args, **kwargs):
        BaseConnector.__init__(self, **kwargs)
        PredicatePushable.__init__(self, **kwargs)
        if ".." in self.dataset or self.dataset[0] in ("\\", "/", "~"):
            # Don't find any datasets which look like path traversal
            raise DatasetNotFoundError(dataset=self.dataset)
        self.decoder = get_decoder(self.dataset)

    def _read_file(self) -> None:
        """
        Reads the dataset file and stores its content in _byte_array attribute.
        """
        if self._byte_array is None:
            with open(self.dataset, mode="br") as file:
                self._byte_array = bytes(file.read())

    def read_dataset(
        self, columns: list = None, predicates: list = None, **kwargs
    ) -> pyarrow.Table:
        """
        Reads the dataset file and decodes it.

        Returns:
            An iterator containing a single decoded pyarrow.Table.
        """
        self._read_file()
        return iter([self.decoder(self._byte_array, projection=columns, selection=predicates)])

    def get_dataset_schema(self) -> RelationSchema:
        """
        Retrieves the schema from the dataset file.

        Returns:
            The schema of the dataset.
        """
        if self.schema is not None:
            return self.schema

        self._read_file()
        self.schema = self.decoder(self._byte_array, just_schema=True)
        return self.schema



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opteryx.components.logical_planner import LogicalPlan
from opteryx.components.logical_planner import LogicalPlanNode


# Context object to carry state
class HeuristicOptimizerContext:
    def __init__(self, tree: LogicalPlan):
        self.node_id = None
        self.parent_nid = None
        self.last_nid = None
        self.pre_optimized_tree = tree
        self.optimized_plan = LogicalPlan()

        # We collect predicates we should be able to push to reads and joins
        self.collected_predicates: list = []

        # We collect column identities so we can push column selection as close to the
        # read as possible, including off to remote systems
        self.collected_identities: set = set()


class OptimizationStrategy:
    def visit(
        self, node: LogicalPlanNode, context: HeuristicOptimizerContext
    ) -> HeuristicOptimizerContext:
        raise NotImplementedError(
            "Visit method must be implemented in OptimizationStrategy classes."
        )

    def complete(self, plan: LogicalPlan, context: HeuristicOptimizerContext) -> LogicalPlan:
        raise NotImplementedError(
            "Complete method must be implemented in OptimizationStrategy classes."
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Heap Sort Node

This is a SQL Query Execution Plan Node.

This node orders a dataset, note that Heap Sort in this instance isn't the heap sort
algorithm, it is an approach where a heap of n items (the limit) is maintained as the
data passes through the operator. Because we are working with chunks, we build a 
temporary batch of n + len(batch) items, sort them and then discard all but n of
the items.

This is faster, particularly when working with large datasets even though we're now
sorting smaller datasets over and over again.

Early testing with a ~10m record dataset:
- sort then limit: 39.2 seconds
- heapsort: 9.8 seconds
"""
import time
from typing import Iterable

import pyarrow
from orso.types import OrsoTypes
from pyarrow import Table

from opteryx.exceptions import ColumnNotFoundError
from opteryx.exceptions import SqlError
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import format_expression
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


class HeapSortNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self.order = config.get("order", [])
        self.limit: int = config.get("limit", -1)

    @property
    def greedy(self):  # pragma: no cover
        return True

    @property
    def config(self):  # pragma: no cover
        return f"LIMIT = {self.limit} ORDER = " + ", ".join(
            [f"{i[0][0].value} {i[1][0:3].upper()}" for i in self.order]
        )

    @property
    def name(self):  # pragma: no cover
        return "Heap Sort"

    def execute(self) -> Iterable:
        morsels = self._producers[0]  # type:ignore
        if isinstance(morsels, Table):
            morsels = (morsels,)

        columns = None
        collected_rows = None
        mapped_order = []

        for morsel in morsels.execute():
            if columns is None:
                columns = Columns(morsel)

                for column_list, direction in self.order:
                    for column in column_list:
                        if (
                            column.node_type == NodeType.LITERAL
                            and column.type == OrsoTypes.INTEGER
                        ):
                            # we have an index rather than a column name, it's a natural
                            # number but the list of column names is zero-based, so we
                            # subtract one
                            column_name = morsel.column_names[int(column.value) - 1]
                            mapped_order.append(
                                (
                                    column_name,
                                    direction,
                                )
                            )
                        else:
                            try:
                                mapped_order.append(
                                    (
                                        columns.get_column_from_alias(
                                            format_expression(column), only_one=True
                                        ),
                                        direction,
                                    )
                                )
                            except ColumnNotFoundError as cnfe:
                                raise ColumnNotFoundError(
                                    f"`ORDER BY` must reference columns as they appear in the `SELECT` clause. {cnfe}"
                                )

            start_time = time.time_ns()

            # add what we've collected before to the table
            if collected_rows:  # pragma: no cover
                self.statistics.morsel_merges += 1
                morsel = pyarrow.concat_tables([collected_rows, morsel], promote_options="none")

            morsel = morsel.sort_by(mapped_order)
            collected_rows = morsel.slice(offset=0, length=self.limit)

            self.statistics.time_ordering += time.time_ns() - start_time

        yield collected_rows



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Node Module

This module contains the Node class, which provides an implementation for dynamic attribute management.

Noteworthy features and design choices:

1. Dynamic Attributes: The Node class allows you to set and get attributes dynamically, storing them in an internal dictionary.
2. Attribute Validation: Attributes starting with an underscore are not allowed.
3. Property Access: The class provides a `properties` method that exposes the internal attributes for external use.
4. Attribute Defaults: When attempting to access an attribute that doesn't exist, the `__getattr__` method will return None.
5. Deep Copy: The `copy` method allows for deep copying of the Node object, preserving the structure and values of all internal attributes.
6. JSON Representation: The `__str__` method returns a JSON representation of the internal attributes, which can be helpful for debugging or serialization.

"""

import copy
from typing import Any
from typing import Dict
from typing import Union


class Node:
    def __init__(self, node_type: Union[str, int, None] = None, **kwargs: Any):
        """
        Initialize a Node with attributes.

        Parameters:
            node_type: str, optional
                The type of the node.
            **kwargs: Any
                Dynamic attributes for the node.
        """
        object.__setattr__(self, "_internal", {})  # Directly set _internal using the base method
        if node_type:
            self._internal["node_type"] = node_type
        for k, v in kwargs.items():
            self._is_valid_key(k)
            self._internal[k] = v

    def _is_valid_key(self, key: str) -> None:
        """
        Check if the key is valid for the node.

        Parameters:
            key: str
                The key to check.
        """
        if key.startswith("_"):
            raise AttributeError("Node cannot have dynamic attributes starting with an underscore")

    @property
    def properties(self) -> Dict[str, Any]:
        """
        Get the internal properties of the Node.

        Returns:
            Dict[str, Any]: The internal properties.
        """
        return self._internal

    def __getattr__(self, name: str) -> Any:
        """
        Retrieve attribute from the internal dictionary or the _identity.

        Parameters:
            name: str
                The name of the attribute to retrieve.

        Returns:
            Any: The attribute value.
        """
        return self._internal.get(name)

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Set attribute in the internal dictionary.

        Parameters:
            name: str
                The name of the attribute.
            value: Any
                The value to set.
        """
        if name == "_internal":
            self._internal = value
        else:
            self._is_valid_key(name)
            if value is None:
                self._internal.pop(name, None)
            else:
                self._internal[name] = value

    def __str__(self) -> str:
        """
        Return a string representation of the Node using JSON serialization.

        Returns:
            str: The JSON string representation.
        """
        import orjson

        return orjson.dumps(self._internal, default=str).decode()

    def __repr__(self) -> str:
        """
        Provide a detailed representation for debugging.

        Returns:
            str: A string representation useful for debugging.
        """
        node_type = str(self._internal.get("node_type", "<unspecified>"))
        if node_type.startswith("LogicalPlanStepType."):
            node_type = node_type[20:]
        return f"<Node type={node_type}>"

    def copy(self) -> "Node":
        """
        Create an independent deep copy of the node.

        Returns:
            Node: The new, independent deep copy.
        """

        def _inner_copy(obj: Any) -> Any:
            """
            Create an independent inner copy of the given object.

            Parameters:
                obj: Any
                    The object to be deep copied.

            Returns:
                Any: The new, independent deep copy.
            """
            try:
                if isinstance(obj, list):
                    return [_inner_copy(item) for item in obj]
                if isinstance(obj, tuple):
                    return tuple(_inner_copy(item) for item in obj)
                if isinstance(obj, set):
                    return {_inner_copy(item) for item in obj}
                if isinstance(obj, dict):
                    return {key: _inner_copy(value) for key, value in obj.items()}
                if hasattr(obj, "copy"):
                    return obj.copy()
                return copy.deepcopy(obj)
            except:
                return obj

        new_node = Node()
        for key, value in self._internal.items():
            new_value = _inner_copy(value)
            setattr(new_node, key, new_value)
        return new_node



"""
This module provides an interface to the sqloxide library, which is responsible for parsing SQL,
restoring the Abstract Syntax Tree (AST), and performing various mutations on expressions and relations.

For more information about sqloxide: https://github.com/wseaton/sqloxide

This module is not from sqloxide, it is written for Opteryx.
"""

from .sqloxide import mutate_expressions
from .sqloxide import mutate_relations
from .sqloxide import parse_sql
from .sqloxide import restore_ast

# Explicitly define the API of this module for external consumers
__all__ = ["parse_sql", "restore_ast", "mutate_expressions", "mutate_relations"]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Distinct Node

This is a SQL Query Execution Plan Node.

This Node eliminates duplicate records.
"""
import time
from typing import Generator

import pyarrow
from pyarrow import concat_tables

from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.third_party.pyarrow_ops import drop_duplicates


class DistinctNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._distinct_on = config.get("on")
        if self._distinct_on:
            self._distinct_on = [col.schema_column.identity for col in self._distinct_on]

    @property
    def config(self):  # pragma: no cover
        return ""

    @property
    def greedy(self):  # pragma: no cover
        return True

    @property
    def name(self):  # pragma: no cover
        return "Distinction"

    def execute(self) -> Generator[pyarrow.Table, None, None]:
        morsels = self._producers[0]  # type:ignore

        start = time.monotonic_ns()
        dropped = drop_duplicates(
            concat_tables(morsels.execute(), promote_options="none"), self._distinct_on
        )
        self.statistics.time_distincting += time.monotonic_ns() - start

        yield dropped



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This implements an interface to Memcached

If we have 10 failures in a row, stop trying to use the cache.
"""

import os
from typing import Union

from orso.tools import single_item_cache

from opteryx.exceptions import MissingDependencyError
from opteryx.managers.kvstores import BaseKeyValueStore

MAXIMUM_CONSECUTIVE_FAILURES: int = 10


@single_item_cache
def _memcached_server(**kwargs):
    """
    Handling connecting to Memcached
    """
    # the server must be set in the environment
    memcached_config = kwargs.get("server", os.environ.get("MEMCACHED_SERVER"))
    if memcached_config is None:
        return None

    # expect either SERVER or SERVER:PORT entries
    memcached_config = memcached_config.split(":")
    if len(memcached_config) == 1:
        # the default memcached port
        memcached_config.append(11211)

    # we need the server and the port
    if len(memcached_config) != 2:
        return None

    try:
        from pymemcache.client import base
    except ImportError as err:
        raise MissingDependencyError(err.name) from err

    try:
        # wait 1 second to try to connect, it's not worthwhile as a cache if it's slow
        cache = base.Client(
            (
                memcached_config[0],
                memcached_config[1],
            ),
            connect_timeout=1,
            timeout=1,
        )
    except Exception as err:
        print("[CACHE] Unable to create remote cache", err)
        cache = None

    return cache


class MemcachedCache(BaseKeyValueStore):
    """
    Cache object
    """

    def __init__(self, **kwargs):
        """
        Parameters:
            servers: string (optional)
                Sets the memcached server and port (server:port). If not provided
                the value will be obtained from the OS environment.
        """
        self._server = _memcached_server(**kwargs)
        if self._server is None:
            self._consecutive_failures: int = MAXIMUM_CONSECUTIVE_FAILURES
        else:
            self._consecutive_failures = 0
        self.hits: int = 0
        self.misses: int = 0
        self.skips: int = 0
        self.errors: int = 0

    def get(self, key: bytes) -> Union[bytes, None]:
        if self._consecutive_failures >= MAXIMUM_CONSECUTIVE_FAILURES:
            self.skips += 1
            return None
        try:
            response = self._server.get(key)
            self._consecutive_failures = 0
            if response:
                self.hits += 1
                return bytes(response)
        except Exception as err:  # pragma: no cover
            self._consecutive_failures += 1
            if self._consecutive_failures >= MAXIMUM_CONSECUTIVE_FAILURES:
                import datetime

                print(
                    f"{datetime.datetime.now()} [CACHE] Disabling remote Memcached cache due to persistent errors ({err})."
                )
            self.errors += 1
            return None

        self.misses += 1
        return None

    def set(self, key: bytes, value: bytes) -> None:
        if self._consecutive_failures < MAXIMUM_CONSECUTIVE_FAILURES:
            self._server.set(key, value)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is a long module, the intent is to have the functions defined in alphabetical order to
help find them
"""
import datetime
import inspect
import re
import sys
import typing
from enum import Enum
from enum import auto

import numpy
from orso.tools import single_item_cache
from pyarrow import compute

CAMEL_TO_SNAKE = re.compile(r"(?<!^)(?=[A-Z])")


def not_implemented(*args, **kwds):
    raise NotImplementedError("Subclasses must implement the _func method.")


@single_item_cache
def get_functions():
    """so we don't have to manually maintain a list of functions, build it using introspection"""
    functions = {}
    members = inspect.getmembers(
        sys.modules[__name__],
        lambda member: inspect.isclass(member) and member.__module__ == __name__,
    )
    for function_name, function_implementation in members:
        if function_name[0] != "_":
            # python CamelCase to SQL SNAKE_CASE
            function_name = CAMEL_TO_SNAKE.sub("_", function_name).upper()
            if function_name.startswith("FUNCTION_"):
                function_name = function_name[9:]
                functions[function_name] = function_implementation
                for alias in function_implementation.aliases:
                    functions[alias] = function_implementation

    return functions


class _Functions:
    def __init__(self):
        self._functions = get_functions()

    def get(self, func):
        """return a function with a given name"""
        implementation = self._functions.get(func)
        if implementation is None:
            return None
        return implementation()

    def suggest(self, func):
        """return the function with the nearest name match"""
        from itertools import permutations

        from opteryx.utils import suggest_alternative

        available_functions = self._functions.keys()

        # try a fuzzy search (typos) first
        suggestion = suggest_alternative(func, available_functions)

        # if it probably wasn't a typo, try rearranging the parts of the function names
        if suggestion is None:
            parts = func.split("_")
            combinations = permutations(parts)
            for combination in combinations:
                candidate = "_".join(combination)
                suggestion = suggest_alternative(candidate, available_functions)
                if suggestion:
                    break

        return suggestion

    def collect(self, full_details: bool = False):
        """return all of the functions"""
        if full_details:
            function_list = []
            for function, implementation in self._functions.items():
                concrete_implementation = implementation()
                if (
                    function.replace("_", "")
                    != concrete_implementation.__class__.__name__[8:].upper()
                ):
                    # we're an alias
                    continue
                args = concrete_implementation.argument_types() or []
                function_definition = {
                    "Function": function,
                    "Description": concrete_implementation.describe(),
                    "Arguments": None if args == [] else ", ".join(arg[4] for arg in args),
                    "Return_Type": ", ".join(concrete_implementation.return_types()),
                    "Function_Type": concrete_implementation.style_name(),
                }
                function_list.append(function_definition)
                for alias in concrete_implementation.aliases:
                    if alias != function:
                        function_definition = {
                            "Function": alias,
                            "Description": f"Alias of `{function}`",
                            "Argumentss": None,
                            "Return_Type": None,
                            "Function_Type": None,
                        }
                        function_list.append(function_definition)
            return function_list
        return list(self._functions.keys())


class _FunctionStyle(Enum):
    # Aggregation functions accept a column of values and return a single value
    AGGREGATION = auto()
    # Elementwise functions accept a set of two or more columns of the same length and return
    # a column of the same length as the inputs
    ELEMENTWISE = auto()
    # Setwise functions accept one column and zero or more fixed values and return a column of
    # the same length as the input
    SETWISE = auto()
    # Constant functions return a single value, usually with no input
    CONSTANT = auto()


class _BaseFunction:
    # The inner function that is executed
    _func = not_implemented
    # The approximate cost to perform the function.
    # This is approximately the time to execute the function 1 million times in 1000s of a second.
    cost: typing.Union[int, None] = None
    # The range of values returnable by this value
    value_range: typing.Union[typing.Tuple[typing.Any], None] = None
    # The type of function
    style: _FunctionStyle = _FunctionStyle.ELEMENTWISE
    # Is the order order maintained by this function
    order_maintained: bool = False
    # Does this functiona have any aliases
    aliases: list = []

    def __call__(self, *args: typing.Any, **kwds: typing.Any) -> typing.Any:
        return self._func(*args, **kwds)

    def describe(self):
        return f"{self.__doc__.strip()}"

    @property
    def signature(self):
        def _render_arg(arg):
            name, _, default, optional, _ = arg
            if not optional:
                return f"{name}:type"
            if default is False:
                return f"[{name}:type]"
            return f"[{name}:type = {default}]"

        args = self.argument_types() or []
        params = "" if args == [] else ", ".join(_render_arg(arg) for arg in args)
        return f"{self.__class__.__name__[8:].upper()}({params}) → {self.return_types()[0]}"

    def name(self):
        return f"{self.__class__.__name__[8:].upper()}"

    def style_name(self):
        if self.style in (_FunctionStyle.ELEMENTWISE, _FunctionStyle.SETWISE):
            return "SCALAR"
        return str(self.style.name)

    def return_types(self):
        from orso.dataframe import TYPE_MAP

        TYPE_MAP[typing.Any] = "INPUT TYPE"

        return_type_hints = typing.get_type_hints(self._func).get("return")

        if return_type_hints is not None:
            if typing.get_origin(return_type_hints) is typing.Union:
                return [
                    TYPE_MAP.get(return_type, "OTHER")
                    for return_type in typing.get_args(return_type_hints)
                ]
            return [TYPE_MAP.get(return_type_hints, "OTHER")]
        raise NotImplementedError(
            f"{self.__class__.__name__.upper()} hasn't specified its return types"
        )

    def argument_types(self):
        from orso.dataframe import TYPE_MAP

        def is_optional_type(param_type):
            origin = typing.get_origin(param_type)
            return origin is typing.Union and type(None) in typing.get_args(param_type)

        func_signature = inspect.signature(self._func)
        func_parameters = func_signature.parameters
        type_hints = typing.get_type_hints(self._func)

        return_value = []
        for arg_name, arg_value in func_parameters.items():
            arg_type = type_hints.get(arg_name)
            default = None
            if isinstance(arg_value.default, (str, int, float, bool)):
                default = arg_value.default

            optional = is_optional_type(arg_type) or default is not None
            argument = (
                arg_name,
                TYPE_MAP.get(arg_type, arg_type),
                default,
                optional,
                f"{arg_value}{'='+str(default) if default else ''}{' OPTIONAL' if optional else ''}",
            )
            return_value.append(argument)
        if return_value == []:
            return_value = None
        return return_value

    def validate_func_parameters(self, *args):
        func_signature = inspect.signature(self._func)
        func_parameters = func_signature.parameters
        type_hints = typing.get_type_hints(self._func)

        # Validate positional arguments
        for i, (arg_name, arg_value) in enumerate(func_parameters.items()):
            arg_type = type_hints.get(arg_name)
            if arg_type and not isinstance(arg_type, type) and not issubclass(arg_type, type):
                raise TypeError(f"Invalid type hint for argument '{arg_name}': {arg_type}")

            if not isinstance(arg_value.default, type) and not issubclass(arg_value.default, type):
                raise TypeError(f"Invalid type hint for argument '{arg_name}': {arg_value.default}")

            if i < len(args) and arg_type:
                if not isinstance(args[i], arg_type):
                    return False

        return True

    def determine_input_values(self):
        """
        Determine the input values for measuring the execution time of the given function.
        """
        parameters = inspect.signature(self._func).parameters
        type_hints = typing.get_type_hints(self._func)

        args = []
        value: typing.Any = None
        for parameter_name, parameter in parameters.items():
            # Check if the parameter has a type hint
            if parameter_name in type_hints:
                parameter_type = type_hints[parameter_name]
                # Optional
                if isinstance(parameter_type.__args__, tuple):
                    parameter_type = parameter_type.__args__[0]
                # Generate a value based on the parameter type (customize as needed)
                if parameter_type == int:
                    value = 10
                elif parameter_type == str:
                    value = "example"
                elif parameter_type == list:
                    value = [1, 2, 3]
                # Add more conditions for other parameter types as necessary
                else:
                    # Handle unsupported parameter types or default to None
                    value = None
                args.append(value)
            else:
                # Handle parameters without type hints or default to None
                args.append(None)

        return args

    def calculate_cost(self):
        """
        You're not meant to call this function, I'm used internally for approximating the
        cost value for this function.

        Cost is roughly the number of nanoseconds to execute the function 1 million times.
        """
        import time

        import numpy

        CYCLES = 50

        args = self.determine_input_values()

        measurements = []
        for cycle in range(CYCLES):
            start = time.monotonic_ns()
            for _ in range(1000000):
                self._func(*args)
            measurements.append((time.monotonic_ns() - start) / 1000000)

        threshold = 3
        mean = numpy.mean(measurements)
        std = numpy.std(measurements)
        # remove outliers - anything 3 standard deviations from the mean
        measurements = [x for x in measurements if abs(x - mean) < threshold * std]
        # 80% of responses were below p80
        p80 = numpy.percentile(measurements, 80, method="nearest")

        return numpy.round(p80 / 50) * 50


class FunctionCurrentTime(_BaseFunction):
    """Return the current system time."""

    style = _FunctionStyle.CONSTANT
    cost = 450
    returns_nulls = False

    def _func(self) -> datetime.time:
        return datetime.datetime.utcnow().time()


class FunctionE(_BaseFunction):
    """Return Euler's number."""

    style = _FunctionStyle.CONSTANT
    cost = 150

    def _func(self) -> float:
        return 2.718281828459045235360287


class FunctionGreatest(_BaseFunction):
    """Return the greatest value in array."""

    style = _FunctionStyle.SETWISE
    cost = 150

    def _func(self, array: list) -> typing.Any:
        return numpy.nanmax(array)


class FunctionLen(_BaseFunction):
    """Return the length of a VARCHAR or ARRAY"""

    style = _FunctionStyle.ELEMENTWISE
    cost = 250
    order_maintained = False
    aliases = ["LENGTH"]

    def _func(self, value: typing.Union[list, str]) -> int:
        return len(value)


class FunctionPhi(_BaseFunction):
    """Return the golden ratio."""

    style = _FunctionStyle.CONSTANT
    cost = 150

    def _func(self) -> float:
        return 1.618033988749894848204586


class FunctionPi(_BaseFunction):
    """Return Pi."""

    style = _FunctionStyle.CONSTANT
    cost = 150

    def _func(self) -> float:
        return 3.141592653589793238462643


class FunctionRound(_BaseFunction):
    """Returns `value` rounded to `places` decimal places."""

    style = _FunctionStyle.SETWISE

    def _func(self, value: typing.List[float], places: int = 0) -> float:
        return compute.round(value, places)  # [#325]


class FunctionVersion(_BaseFunction):
    """Return the version of the query engine."""

    style = _FunctionStyle.CONSTANT
    cost = 350

    def _func(self) -> str:
        import opteryx

        return opteryx.__version__


FUNCTIONS = _Functions()

if __name__ == "__main__":  # pragma: no cover
    import orso

    function_table = orso.DataFrame(FUNCTIONS.collect(True))
    print(function_table)

    func = FUNCTIONS.get("ROUND")
    print(func)
    print(func.style)
    print(func.return_types())
    print(func.argument_types())
    print(func.signature)

    import time

    print(f"Did you mean {FUNCTIONS.suggest('time_current')}?")
#    for f in FUNCTIONS.collect(False):
#        func = FUNCTIONS.get(f)
#        print(f"running {f} 1 million times took {func.calculate_cost()}")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
~~~
                      ┌───────────┐
                      │   USER    │
         ┌────────────┤           ◄────────────┐
         │SQL         └───────────┘            │
  ───────┼─────────────────────────────────────┼──────
         │                                     │
   ┌─────▼─────┐                               │
   │ SQL       │                               │
   │  Rewriter │                               │
   └─────┬─────┘                               │
         │SQL                                  │Plan
   ┌─────▼─────┐                         ┌─────┴─────┐
   │           │                         │           │
   │ Parser    │                         │ Executor  │
   └─────┬─────┘                         └─────▲─────┘
         │AST                                  │Plan
   ┌─────▼─────┐      ┌───────────┐      ┌─────┴─────┐
   │ AST       │      │           │Stats │Cost-Based │
   │ Rewriter  │      │ Catalogue ├──────► Optimizer │
   └─────┬─────┘      └─────┬─────┘      └─────▲─────┘
         │AST               │Schemas           │Plan
   ┌─────▼─────┐      ┌─────▼─────┐      ┌─────┴─────┐
   │ Logical   │ Plan │           │ Plan │ Heuristic │
   │   Planner ├──────► Binder    ├──────► Optimizer │
   └───────────┘      └───────────┘      └───────────┘
~~~
"""

from typing import Iterable
from typing import Union

from opteryx import config

PROFILE_LOCATION = config.PROFILE_LOCATION


def query_planner(operation: str, parameters: Union[Iterable, None], connection, qid: str):
    import orjson

    from opteryx.components.ast_rewriter import do_ast_rewriter
    from opteryx.components.binder import do_bind_phase
    from opteryx.components.cost_based_optimizer import do_cost_based_optimizer
    from opteryx.components.heuristic_optimizer import do_heuristic_optimizer
    from opteryx.components.logical_planner import LogicalPlan
    from opteryx.components.logical_planner import do_logical_planning_phase
    from opteryx.components.sql_rewriter import do_sql_rewrite
    from opteryx.components.temporary_physical_planner import create_physical_plan
    from opteryx.exceptions import SqlError
    from opteryx.models import QueryProperties
    from opteryx.third_party import sqloxide

    # SQL Rewriter removes whitespace and comments, and extracts temporal filters
    clean_sql, temporal_filters = do_sql_rewrite(operation)
    params = [p for p in parameters or []]

    profile_content = operation + "\n\n"
    # Parser converts the SQL command into an AST
    try:
        parsed_statements = sqloxide.parse_sql(clean_sql, dialect="mysql")
    except ValueError as parser_error:
        raise SqlError(parser_error) from parser_error
    # AST Rewriter adds temporal filters and parameters to the AST
    parsed_statements = do_ast_rewriter(
        parsed_statements,
        temporal_filters=temporal_filters,
        paramters=params,
        connection=connection,
    )

    logical_plan: LogicalPlan = None
    ast: dict = {}

    # Logical Planner converts ASTs to logical plans
    for logical_plan, ast, ctes in do_logical_planning_phase(parsed_statements):  # type: ignore
        # check user has permission for this query type
        query_type = next(iter(ast))
        if query_type not in connection.permissions:
            from opteryx.exceptions import PermissionsError

            raise PermissionsError(
                f"User does not have permission to execute '{query_type}' queries."
            )

        profile_content += (
            orjson.dumps(logical_plan.depth_first_search(), option=orjson.OPT_INDENT_2).decode()
            + "\n\n"
        )
        profile_content += logical_plan.draw() + "\n\n"

        heuristic_optimized_plan = do_heuristic_optimizer(logical_plan)

        # The Binder adds schema information to the logical plan
        bound_plan = do_bind_phase(
            heuristic_optimized_plan,
            connection=connection.context,
            qid=qid,
            # common_table_expressions=ctes,
        )

        optimized_plan = do_cost_based_optimizer(bound_plan)

        # before we write the new optimizer and execution engine, convert to a V1 plan
        query_properties = QueryProperties(qid=qid, variables=connection.context.variables)
        physical_plan = create_physical_plan(optimized_plan, query_properties)
        yield physical_plan



from typing import Union

from opteryx import config
from opteryx.exceptions import InvalidConfigurationError
from opteryx.managers.kvstores import BaseKeyValueStore

MAX_CACHEABLE_ITEM_SIZE = config.MAX_CACHEABLE_ITEM_SIZE
MAX_CACHE_EVICTIONS_PER_QUERY = config.MAX_CACHE_EVICTIONS_PER_QUERY
MAX_LOCAL_BUFFER_CAPACITY = config.MAX_LOCAL_BUFFER_CAPACITY


class CacheManager:
    """
    Manages cache behavior for Opteryx, including the cache store, item size, eviction policies, etc.

    Parameters:
        cache_backend: Union[BaseKeyValueStore, None]
            The cache storage to use.
        max_cacheable_item_size: int
            The maximum size a single item in the cache can occupy.
        max_evictions_per_query: int
            The number of items to evict from cache per query.
        max_local_buffer_capacity: int
            The maximum number of items to store in the BufferPool.
    """

    def __init__(
        self,
        cache_backend: Union[BaseKeyValueStore, None] = None,
        max_cacheable_item_size: Union[int, None] = MAX_CACHEABLE_ITEM_SIZE,
        max_evictions_per_query: Union[int, None] = MAX_CACHE_EVICTIONS_PER_QUERY,
        max_local_buffer_capacity: int = MAX_LOCAL_BUFFER_CAPACITY,
    ):
        if cache_backend is not None and not isinstance(cache_backend, BaseKeyValueStore):
            raise InvalidConfigurationError(
                config_item="cache_backend",
                provided_value=str(type(cache_backend)),
                valid_value_description="Instance of BaseKeyValueStore",
            )

        if max_cacheable_item_size is not None and (
            not isinstance(max_cacheable_item_size, int) or max_cacheable_item_size <= 0
        ):
            raise InvalidConfigurationError(
                config_item="max_cacheable_item_size",
                provided_value=str(max_cacheable_item_size),
                valid_value_description="A number greater than zero",
            )

        if max_evictions_per_query is not None and (
            not isinstance(max_evictions_per_query, int) or max_evictions_per_query <= 0
        ):
            raise InvalidConfigurationError(
                config_item="max_evictions_per_query",
                provided_value=str(max_evictions_per_query),
                valid_value_description="A number greater than zero",
            )

        if not isinstance(max_local_buffer_capacity, int) or max_local_buffer_capacity <= 0:
            raise InvalidConfigurationError(
                config_item="max_local_buffer_capacity",
                provided_value=str(max_local_buffer_capacity),
                valid_value_description="A number greater than zero",
            )

        self.cache_backend = cache_backend
        self.max_cacheable_item_size = max_cacheable_item_size
        self.max_evictions_per_query = max_evictions_per_query
        self.max_local_buffer_capacity = max_local_buffer_capacity



from .cache_manager import CacheManager
from .memcached import MemcachedCache
from .memory import MemoryCache
from .redis import RedisCache

__all__ = ("CacheManager", "MemcachedCache", "MemoryCache", "RedisCache")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The 'direct disk' connector provides the reader for when a dataset is
given as a folder on local disk
"""

import os
from typing import Dict
from typing import List

import pyarrow
from orso.schema import RelationSchema
from orso.tools import single_item_cache
from orso.types import OrsoTypes

from opteryx.connectors.base.base_connector import BaseConnector
from opteryx.connectors.capabilities import Cacheable
from opteryx.connectors.capabilities import Partitionable
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.exceptions import DatasetNotFoundError
from opteryx.exceptions import EmptyDatasetError
from opteryx.exceptions import UnsupportedFileTypeError
from opteryx.utils.file_decoders import TUPLE_OF_VALID_EXTENSIONS
from opteryx.utils.file_decoders import get_decoder

OS_SEP = os.sep


class DiskConnector(BaseConnector, Cacheable, Partitionable, PredicatePushable):
    __mode__ = "Blob"

    PUSHABLE_OPS: Dict[str, bool] = {
        "Eq": True,
        "NotEq": True,
        "Gt": True,
        "GtEq": True,
        "Lt": True,
        "LtEq": True,
    }

    PUSHABLE_TYPES = {OrsoTypes.BOOLEAN, OrsoTypes.DOUBLE, OrsoTypes.INTEGER, OrsoTypes.VARCHAR}

    def __init__(self, **kwargs):
        """
        Initialize the DiskConnector, which reads datasets directly from disk.

        Parameters:
            kwargs: Dict
                Arbitrary keyword arguments.
        """
        BaseConnector.__init__(self, **kwargs)
        Partitionable.__init__(self, **kwargs)
        Cacheable.__init__(self, **kwargs)
        PredicatePushable.__init__(self, **kwargs)

        self.dataset = self.dataset.replace(".", OS_SEP)
        self.cached_first_blob = None  # Cache for the first blob in the dataset

    def read_blob(self, *, blob_name, **kwargs) -> bytes:
        """
        Read a blob (binary large object) from disk.

        We're using the low-level read, on the whole it's about 5% faster - not
        much faster considering the effort to bench-mark different disk access
        methods, but as one of the slowest parts of the system we wanted to find
        if there was a faster way.

        Parameters:
            blob_name: str
                The name of the blob file to read.
            kwargs: Dict
                Arbitrary keyword arguments.

        Returns:
            The blob as bytes.
        """
        file_descriptor = os.open(blob_name, os.O_RDONLY)
        try:
            return os.read(file_descriptor, os.path.getsize(blob_name))
        finally:
            os.close(file_descriptor)

    @single_item_cache
    def get_list_of_blob_names(self, *, prefix: str) -> List[str]:
        """
        List all blob files in the given directory path.

        Parameters:
            prefix: str
                The directory path.

        Returns:
            A list of blob filenames.
        """
        return sorted(
            os.path.join(root, file)
            for root, _, files in os.walk(prefix)
            for file in files
            if file.endswith(TUPLE_OF_VALID_EXTENSIONS)
        )

    def read_dataset(
        self, columns: list = None, predicates: list = None, just_schema: bool = False, **kwargs
    ) -> pyarrow.Table:
        """
        Read the entire dataset from disk.

        Yields:
            Each blob's content as a PyArrow Table.
        """
        blob_names = self.partition_scheme.get_blobs_in_partition(
            start_date=self.start_date,
            end_date=self.end_date,
            blob_list_getter=self.get_list_of_blob_names,
            prefix=self.dataset,
        )

        for blob_name in blob_names:
            try:
                decoder = get_decoder(blob_name)
                blob_bytes = self.read_blob(blob_name=blob_name, statistics=self.statistics)
                yield decoder(
                    blob_bytes, projection=columns, selection=predicates, just_schema=just_schema
                )
            except UnsupportedFileTypeError:
                pass  # Skip unsupported file types
            except pyarrow.ArrowInvalid:
                self.statistics.unreadable_data_blobs += 1

    def get_dataset_schema(self) -> RelationSchema:
        """
        Retrieve the schema of the dataset either from the metastore or infer it from the first blob.

        Returns:
            The schema of the dataset.
        """
        self.schema = self.read_schema_from_metastore()
        if self.schema:
            return self.schema

        self.schema = next(self.read_dataset(just_schema=True), None)

        if self.schema is None:
            if os.path.isdir(self.dataset):
                raise EmptyDatasetError(dataset=self.dataset.replace("/", "."))
            raise DatasetNotFoundError(dataset=self.dataset)

        return self.schema
